(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file analyze.ml contains the main functions of the analyzer
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Boolean ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Coercion ;;
open Matrix ;;
open Semiformal ;;
open Parser ;;
open Widenings ;;
open Slice ;;
open Environments ;;


(*****************************************************************************)
(*                            Main analyze loop                              *)
(*****************************************************************************)				

let rec removeUndefinedCall expr tEnv vEnv = match expr with
	UndefinedCall(id,ael,l) -> if ((List.length ael)=1) then
									if (isVArray id vEnv) then 
										(if ((getInt (getArrayLength (getEnv id tEnv)))=1) then 
											ArrayElt(id,Cst(IVal(I32(Int32.one),I32(Int32.one)),l),List.hd ael,l)
										 else    
											let d = getArrayLineLength (getEnv id tEnv) in
											let a1 = List.hd ael in
												ArrayBlock(id,a1,Cst(IVal(I32(Int32.one),I32(Int32.one)),l),
															  a1,Cst(IVal(d,d),l),l)
										)
									else 
										let vl = List.map (fun e -> (Cst(analyzeAExpr e tEnv vEnv,l))) ael
										in setUpFunCall id vl l
								else (* ((List.length ael)>1) *)
									if ((List.length ael)=2) then
										if (isVArray id vEnv) then
											let a1 = List.hd ael in
											let a2 = List.hd (List.tl ael) in
												ArrayElt(id,a1,a2,l)
										else
											let vl = List.map (fun e -> (Cst(analyzeAExpr e tEnv vEnv,l))) ael
											in setUpFunCall id vl l
									else
										let vl = List.map (fun e -> (Cst(analyzeAExpr e tEnv vEnv,l))) ael
										in setUpFunCall id vl l
| ArrayExpr(a,_) -> 				analyzeArray a tEnv vEnv 
| Plus(e1,e2,l) -> 					Plus(removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,l)	
| Minus(e1,e2,l) ->					Minus(removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,l)
| Times(e1,e2,l) -> 				Times(removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,l)
| Div(e1,e2,l) -> 					Div(removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,l)
| UMinus(e1,l) ->					UMinus(removeUndefinedCall e1 tEnv vEnv,l)	
| Sqrt(e1,l) ->						Sqrt(removeUndefinedCall e1 tEnv vEnv,l)	
| Abs(e1,l) ->						Abs(removeUndefinedCall e1 tEnv vEnv,l)	
| Min(e1,l) ->						Min(removeUndefinedCall e1 tEnv vEnv,l)	
| Max(e1,l) ->						Max(removeUndefinedCall e1 tEnv vEnv,l)	
| Size(e1,l) ->						Size(removeUndefinedCall e1 tEnv vEnv,l)	
| Cast(t,e1,l) -> 					Cast(t,removeUndefinedCall e1 tEnv vEnv,l)	
| ArrayElt(s,e1,e2,l)-> 			ArrayElt(s,removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,l)
| CreateMatrix(el,m,l)->	 		CreateMatrix(List.map (fun x -> (removeUndefinedCall x tEnv vEnv)) el,m,l) 
| ArrayBlock(s,e1,e2,e3,e4,l) -> 	ArrayBlock(s,removeUndefinedCall e1 tEnv vEnv,removeUndefinedCall e2 tEnv vEnv,removeUndefinedCall e3 tEnv vEnv,removeUndefinedCall e4 tEnv vEnv,l)
| FunCall(s,el,l) -> 				FunCall(s,List.map (fun x -> (removeUndefinedCall x tEnv vEnv)) el,l)
| _ -> expr


and expr2VE expr tEnv vEnv = match expr with
	Cst(c,_) ->					 	VECst(c)
| 	Cast(t,e,_) -> 					VECast(t,expr2VE e tEnv vEnv)
| 	Id(v,_) -> 						VEId(v,1)
| 	ArrayElt(a,e1,e2,_) -> 			VEArrayElt(a,expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	ArrayBlock(a,e1,e2,e3,e4,_) -> 	VEArrayBlock(a,expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv,expr2VE e3 tEnv vEnv,expr2VE e4 tEnv vEnv)
| 	CreateMatrix(el,t,_) -> 		VECreateMatrix(List.map (fun x -> expr2VE x tEnv vEnv) el,t)
| 	Plus(e1,e2,_) -> 				VEPlus(expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	Minus(e1,e2,_) -> 				VEMinus (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	Times(e1,e2,_) -> 				VETimes (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	DotTimes(e1,e2,_) -> 			VEDotTimes (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	DotDiv(e1,e2,_) -> 				VEDotDiv (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	Div(e1,e2,_) -> 				VEDiv (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	Power(e1,e2,_) -> 				VEPower (expr2VE e1 tEnv vEnv,expr2VE e2 tEnv vEnv)
| 	UMinus(e,_) -> 					VEUMinus (expr2VE e tEnv vEnv) 
| 	Sqrt(e,_) -> 					VESqrt (expr2VE e tEnv vEnv) 
|   Abs(e,_) -> 					VEAbs(expr2VE e tEnv vEnv)
|   Min(e,_) -> 					VEMin(expr2VE e tEnv vEnv)
|   Max(e,_) -> 					VEMax(expr2VE e tEnv vEnv)
|   Size(e,_) -> 					VESize(expr2VE e tEnv vEnv)
|	FunCall(f,el,_) -> 				VEFunCall(f,List.map (fun x -> expr2VE x tEnv vEnv) el)
|	ArrayExpr(_) ->					raise (Error ("expr2VE","arrayexpr should not be converted to veexpr")) 
|	UndefinedCall(_) ->				expr2VE (removeUndefinedCall expr tEnv vEnv) tEnv vEnv


and linearize e x notYetX tEnv vEnv = match e with
	Id(id',l) ->					(match x with
										Id(id,_) ->	if ((String.compare id id')=0) then 
														if (notYetX) then (e,false) else (Cst(getEnv id vEnv,l),false)
													else (Cst(getEnv id' vEnv,l),notYetX)
									| _ -> (Cst(getEnv id' vEnv,l),notYetX)
									)
|	ArrayElt(id,e1,e2,l) ->			(match x with
										ArrayElt(_) ->	if (e=x) then 
															if (notYetX) then (e,false) else 
																(Cst(analyzeAExpr e tEnv vEnv,l),false)
														else
															(Cst(analyzeAExpr e tEnv vEnv,l),notYetX)
									| _ -> (Cst(analyzeAExpr e tEnv vEnv,l),notYetX)
									)
|	Cst(c,_) ->						(e,notYetX)
|	Plus(e1,e2,l) ->				let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valPlus v1 v2,l),notYetX'')
										|	_ -> (Plus(e1',e2',l),notYetX'')
									)
|	Minus(e1,e2,l) ->				let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valMinus v1 v2,l),notYetX'')
										|	_ -> (Minus(e1',e2',l),notYetX'')
									)
|	Times(e1,e2,l) ->				let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valMult v1 v2,l),notYetX'')
										|	_ -> (Times(e1',e2',l),notYetX'')
									)
|	DotTimes(e1,e2,l) ->			let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valDotMult v1 v2,l),notYetX'')
										|	_ -> (DotTimes(e1',e2',l),notYetX'')
									)
|	Div(e1,e2,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valDiv v1 v2,l),notYetX'')
										|	_ -> (Div(e1',e2',l),notYetX'')
									)
|	DotDiv(e1,e2,l) ->				let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv in
									let (e2',notYetX'') = linearize e2 x notYetX' tEnv vEnv in
										(match (e1',e2') with
											(Cst(v1,_),Cst(v2,_)) -> (Cst(valDotDiv v1 v2,l),notYetX'')
										|	_ -> (DotDiv(e1',e2',l),notYetX'')
									)
|	Power(_) ->						raise (Error ("linearize","??? Operator '^' not supported in comparisons."))
|	UMinus(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (UMinus(e1',l),notYetX')				
|	Sqrt(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (Sqrt(e1',l),notYetX')				
|	Abs(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (Abs(e1',l),notYetX')
|	Min(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (Min(e1',l),notYetX')
|	Max(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (Max(e1',l),notYetX')
|	Size(e1,l) ->					let (e1',notYetX') = linearize e1 x notYetX tEnv vEnv
									in (Size(e1',l),notYetX')
|	Cast(_) ->						raise (Error ("linearize","??? Cast not allowed in condition."))
|	CreateMatrix(el,m,l)-> 			(e,notYetX) 
|	ArrayBlock(_,e1,e2,e3,e4,_) -> 	(e,notYetX)	
|	ArrayExpr(a,_) ->				(e,notYetX)
|	FunCall(_,ael,l) ->				(Cst(analyzeAExpr e tEnv vEnv,l),notYetX)
|	UndefinedCall(_,ael,l) ->		(Cst(analyzeAExpr e tEnv vEnv,l),notYetX)


and solve id e op rhs tEnv vEnv = match e with
	Cst(_) ->					analyzeAExpr id tEnv vEnv
|	Id(id',_) ->				if ((String.compare (getId id) id')=0) then
									solveCmp (getEnvComplete id' vEnv ValBottom) op (analyzeAExpr rhs tEnv vEnv)
								else
									getEnvComplete (getId id) vEnv ValBottom
|	ArrayElt(_) -> 				if (id=e) then
									let v = analyzeAExpr e tEnv vEnv
									in solveCmp v op (analyzeAExpr rhs tEnv vEnv)
								else
									analyzeAExpr id tEnv vEnv
|	Plus(Cst(c,l),e',l') -> 	(* c+e < r <=> e < r-c *)
								solve id e' op (Minus(rhs,Cst(c,l),l')) tEnv vEnv
|	Plus(e',Cst(c,l),l') -> 	solve id e' op (Minus(rhs,Cst(c,l),l')) tEnv vEnv
|	Plus(UMinus(Cst(c,l),_),e',l') -> 	
								(* -c+e < r <=> e < r+c *)
								solve id e' op (Plus(rhs,Cst(c,l),l')) tEnv vEnv
|	Plus(e',UMinus(Cst(c,l),_),l') -> 	
								solve id e' op (Plus(rhs,Cst(c,l),l')) tEnv vEnv
|	Plus(UMinus(e',_),UMinus(e'',_),l') -> (* -e' - e'' < r <=> e'+e'' > -r *)
								solve id (Plus(e',e'',l')) (negRelation op) (UMinus(rhs,l')) tEnv vEnv
|	Plus(_)	->					raise (Error ("solve","??? Impossible case plus"))
|	Minus(Cst(c,l),e',l') -> 	(* c-e < r <=> -e < r-c <=> e >= c-r *)
								solve id e' (negRelation op) (Minus(Cst(c,l),rhs,l')) tEnv vEnv
								(* e-c < r <=> e < r+c   *)
|	Minus(e',Cst(c,l),l') -> 	solve id e' op (Plus(rhs,Cst(c,l),l')) tEnv vEnv
|	Minus(e',UMinus(Cst(c,l),l'),_) -> 	
								solve id (Plus(e',Cst(c,l),l')) op rhs tEnv vEnv
|	Minus(UMinus(Cst(c,l),l'),e',_) -> 	
								(* -c-e < r <=> c+e > r <=> e > r-c *)
								solve id e' (negRelation op) (Minus(rhs,Cst(c,l),l')) tEnv vEnv
|	Minus(UMinus(e',_),UMinus(e'',_),l') -> (* -e' - -e'' < r <=> -e'+e'' < r *)
								solve id (Minus(e'',e',l')) op rhs tEnv vEnv
|	Minus(_) ->					raise (Error ("solve","??? Impossible case minus"))
|	Times(Cst(c,l),e',l') -> 	if (isValNonNegative c) then
									solve id e' op (Div(rhs,Cst(c,l),l')) tEnv vEnv
								else
									if (isValNonPositive c) then
										solve id e' (negRelation op) (Div(rhs,Cst(c,l),l')) tEnv vEnv
									else
										let resPos = solve id e' op (Div(rhs,Cst(valMeet c (FVal(!fe128ZeroInfty)),l),l')) tEnv vEnv in
										let resNeg = solve id e' (negRelation op) (Div(rhs,Cst(valMeet c (FVal(!fe128MInftyZero)),l),l')) tEnv vEnv in
											valJoin resPos resNeg
|	Times(e',Cst(c,l),l') -> 	if (isValNonNegative c) then
									solve id e' op (Div(rhs,Cst(c,l),l')) tEnv vEnv
								else
									if (isValNonPositive c) then
										solve id e' (negRelation op) (Div(rhs,Cst(c,l),l')) tEnv vEnv
									else
										let resPos = solve id e' op (Div(rhs,Cst(valMeet c (FVal(!fe128ZeroInfty)),l),l')) tEnv vEnv in
										let resNeg = solve id e' (negRelation op) (Div(rhs,Cst(valMeet c (FVal(!fe128MInftyZero)),l),l')) tEnv vEnv in
											valJoin resPos resNeg											
|	Times(UMinus(Cst(c,l),_),e',l') -> 
								(* -c * e < r <=> ce > -r   *)
								solve id (Times(Cst(c,l),e',l')) (negRelation op) (UMinus(rhs,l')) tEnv vEnv
|	Times(e',UMinus(Cst(c,l),_),l') -> 
								solve id (Times(e',Cst(c,l),l')) (negRelation op) (UMinus(rhs,l')) tEnv vEnv			
|	Times(UMinus(e',_),UMinus(e'',_),l') -> 
								solve id (Times(e',e'',l')) op rhs tEnv vEnv	
|	Times(_) ->					raise (Error ("solve","??? Impossible case times"))											
|	Div(e',Cst(c,l),l') ->	 	if (isValNonNegative c) then
									solve id e' op (Times(rhs,Cst(c,l),l')) tEnv vEnv
								else
									if (isValNonPositive c) then
										solve id e' (negRelation op) (Times(rhs,Cst(c,l),l')) tEnv vEnv
									else
										let resPos = solve id e' op (Times(rhs,Cst(valMeet c (FVal(!fe128ZeroInfty)),l),l')) tEnv vEnv in
										let resNeg = solve id e' (negRelation op) (Times(rhs,Cst(valMeet c (FVal(!fe128MInftyZero)),l),l')) tEnv vEnv in
											valJoin resPos resNeg														
|	Div(Cst(c,l),e',l') ->	 	if (isValNonNegative c) then
									(* c/e < r <=> 1/e < r/c <=> e > c/r  (c >= 0) *)
									solve id e' (negRelation op) (Div(Cst(c,l),rhs,l')) tEnv vEnv
								else
									(* c/e < r <=> 1/e > r/c <=> e < c/r  (c <= 0) *)
									if (isValNonPositive c) then
										solve id e' op (Div(Cst(c,l),rhs,l')) tEnv vEnv
									else
										let resPos = solve id e' (negRelation op) (Div(Cst(valMeet c (FVal(!fe128ZeroInfty)),l),rhs,l')) tEnv vEnv in
										let resNeg = solve id e' op (Div(Cst(valMeet c (FVal(!fe128MInftyZero)),l),rhs,l')) tEnv vEnv in
											valJoin resPos resNeg																									
|	Div(UMinus(Cst(c,l),_),e',l') -> 
								(* -c * e < r <=> ce > -r   *)
								solve id (Div(Cst(c,l),e',l')) (negRelation op) (UMinus(rhs,l')) tEnv vEnv
|	Div(e',UMinus(Cst(c,l),_),l') -> 
								solve id (Div(e',Cst(c,l),l')) (negRelation op) (UMinus(rhs,l')) tEnv vEnv			
|	Div(UMinus(e',_),UMinus(e'',_),l') -> 
								solve id (Div(e',e'',l')) op rhs tEnv vEnv	
|	Div(_) ->					raise (Error ("solve","??? Impossible case div"))																
|	UMinus(e',l') -> 			solve id e' (negRelation op) rhs tEnv vEnv
| 	_ ->						raise (Error ("solve","??? Too complex condition."))
												
											
and analyzeAExpr expr tEnv vEnv =
	let expr' = removeUndefinedCall expr tEnv vEnv in
		match expr' with
			FunCall(id,vel,_) ->	if (!noFunCallYet) then
										let _ = noFunCallYet := false in
										let res = sliceFunCall id vel tEnv vEnv in
										let _ = noFunCallYet := true in res
									else
			 							let vl = List.map (fun x -> (analyzeAExpr x tEnv vEnv)) vel in
										let (tEnv',vEnv') = buildFunCallEnv id vl tEnv vEnv in
										let Fun(_,outputs,body,_) = getEnv id !funEnv in
										let (_,vEnv'') = analyzeCmd body tEnv' vEnv' in
											buildOutputVector outputs vEnv'' []  	
		| 	_ -> 					let initVE = expr2VE expr' tEnv vEnv in
									let vInitVE = valueVESlice initVE tEnv vEnv !slicePrec in 
									let veList = evalFormal expr' vEnv !formalExprMaxHeight tEnv vEnv in
									let (bestVe,bestVal) = selectComplex veList vEnv tEnv initVE vInitVE 
									in
										if ((bestVal=ValBottom)||(bestVal=ValTop)||(valErrLt vInitVE bestVal)) then vInitVE
										else bestVal


and bExprFPStep expr varList op tEnv vEnv res = match varList with
	[] -> res
|	id::ids ->	let (expr',_) = linearize expr (Id(id,Lab(0))) true tEnv vEnv in
				let zero = typedZero (getEnv id vEnv) in
				let v = solve (Id(id,Lab(0))) expr' op zero tEnv vEnv
				in bExprFPStep expr ids op tEnv vEnv (setEnv res id v)

					
and bExprFP expr varList op tEnv vEnv = 
	let vEnv1 = copyEnv vEnv in
	let vEnv2 = bExprFPStep expr varList op tEnv vEnv vEnv1 in
		if (eqEnv vEnv2 vEnv) then vEnv2 else bExprFP expr varList op tEnv vEnv2

										
and analyzeBExpr expr tEnv vEnv = match expr with
	Eq(e1,e2,l) -> 	let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_EQ tEnv env in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_NEQ tEnv env in
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_EQ (analyzeAExpr e2 tEnv' vEnv') in
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	Neq(e1,e2,l) -> let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_NEQ tEnv env in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_EQ tEnv env in		
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in					
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_NEQ (analyzeAExpr e2 tEnv' vEnv') in					
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	Lt(e1,e2,l) -> 	let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_LT tEnv (copyEnv env) in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_GTE tEnv (copyEnv env) in	
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_LT (analyzeAExpr e2 tEnv' vEnv') in	
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	Lte(e1,e2,l) -> let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_LTE tEnv env in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_GT tEnv env in	
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_LTE (analyzeAExpr e2 tEnv' vEnv') in					
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	Gt(e1,e2,l) -> 	let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_GT tEnv env in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_LTE tEnv env in
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_GT (analyzeAExpr e2 tEnv' vEnv') in					
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	Gte(e1,e2,l) -> let varList = (getVarList e1 (getVarList e2 [])) in
					let env = buildCondEnv varList vEnv in
					let trueEnv = bExprFP (Minus(e1,e2,l)) varList Relation_GTE tEnv env in
					let falseEnv = bExprFP (Minus(e1,e2,l)) varList Relation_LT tEnv env in		
					let vEnv' = completeEnv vEnv varList in
					let tEnv' = completeTEnv tEnv varList in		
					let tt = evalCmp (analyzeAExpr e1 tEnv' vEnv') Relation_GTE (analyzeAExpr e2 tEnv' vEnv') in					
						(envMeetAux trueEnv vEnv,envMeetAux falseEnv vEnv,tt)
|	And(e1,e2,_) -> let (trueEnv1,falseEnv1,tt1) = analyzeBExpr e1 tEnv vEnv in
					let (trueEnv2,falseEnv2,tt2) = analyzeBExpr e2 tEnv vEnv in
						(envMeet trueEnv1 trueEnv2,envJoin falseEnv1 falseEnv2,bAnd tt1 tt2)
|	Or(e1,e2,_) -> 	let (trueEnv1,falseEnv1,tt1) = analyzeBExpr e1 tEnv vEnv in
					let (trueEnv2,falseEnv2,tt2) = analyzeBExpr e2 tEnv vEnv in
						(envJoin trueEnv1 trueEnv2,envMeet falseEnv1 falseEnv2,bOr tt1 tt2)
|	Xor(e1,e2,_) ->	let (trueEnv1,falseEnv1,tt1) = analyzeBExpr e1 tEnv vEnv in
					let (trueEnv2,falseEnv2,tt2) = analyzeBExpr e2 tEnv vEnv in
						(envJoin (envMeet trueEnv1 falseEnv2) (envMeet trueEnv2 falseEnv1),
						 envJoin (envMeet falseEnv1 falseEnv2) (envMeet trueEnv1 trueEnv2),bXor tt1 tt2)
|	Not(e1,_) ->	let (trueEnv1,falseEnv1,tt) = analyzeBExpr e1 tEnv vEnv in
						(falseEnv1,trueEnv1,bNot tt)
						
		
and analyzeArray a tEnv vEnv = 
	let y = Array.length a in
	let x = Array.length a.(0) in
	let res = Array.make_matrix y x ValBottom in
	let _ =
		for i=0 to (y-1) do
			for j=0 to (x-1) do
				res.(i).(j) <- analyzeAExpr a.(i).(j) tEnv vEnv 
			done
		done
	in Cst(AVal(res),Lab(0)) 
	
						
and analyzeCmd c tEnv vEnv = match c with
	Nop -> 							(tEnv,vEnv)
|	VarDef(id,typ,_) ->				(setEnv tEnv id typ,vEnv) 
| 	Assign(id,e,Lab(line)) ->		let _ = currentLine := line in
									let v = analyzeAExpr e tEnv vEnv
									in (match (typeOf v) with
									      Type_Fixed(_,_) -> (setEnv tEnv id (maxType (typeOf v) (getEnvComplete id tEnv UndefinedType)),setEnv vEnv id v) 
									    | _ -> (setEnv tEnv id (typeOf v),setEnv vEnv id v)
									   )
| 	AssignBool(id,e,Lab(line)) ->	let _ = currentLine := line in
									let (_,_,v) = analyzeBExpr e tEnv vEnv in
									let v' = (match v with
												B(true) -> !i32One
											 |  B(false) -> !i32Zero
											 |  BTop -> (I32(Int32.zero),I32(Int32.one))
										 	 |  BBottom -> (IBottom,IBottom)
											 )
									in (setEnv tEnv id Type_Int_32,setEnv vEnv id (IVal(v')))
|	AssignArrayElt(id,y,x,e,Lab(line)) ->	
									let _ = currentLine := line in
									let v = analyzeAExpr e tEnv vEnv in
									let a = getArray (getEnv id vEnv) in
									let (y1,y2) = iVal2PairOfInt (analyzeAExpr y tEnv vEnv) vEnv in
									let (x1,x2) = iVal2PairOfInt (analyzeAExpr x tEnv vEnv) vEnv  
									in if ((Array.length a)=1) then 
											let _ = 
												for k=(x1-1) to (x2-1) do
													a.(0).(k) <- cast v (maxType (typeOf a.(0).(k)) (typeOf v)) 
												done  
											in (tEnv,setEnv vEnv id (AVal(a))) 
										else										
											if ((Array.length a.(0))=1) then 
												let _ = 
													for k=(x1-1) to (x2-1) do
														a.(k).(0) <- cast v (maxType (typeOf a.(k).(0)) (typeOf v))
													done  
												in (tEnv,setEnv vEnv id (AVal(a)))
											else
												let _ = 
													for l=(y1-1) to (y2-1) do
														for k=(x1-1) to (x2-1) do
															a.(l).(k) <- cast v (maxType (typeOf a.(l).(k)) (typeOf v))
														done
													done  
												in (tEnv,setEnv vEnv id (AVal(a))) 
| 	AssignArrayBlock(id,y,x,y',x',e,Lab(line)) -> 
									let _ = currentLine := line in
									let v = analyzeAExpr e tEnv vEnv in
									let a = getArray (getEnv id vEnv) in
									let (y1,y2) = iVal2PairOfInt (analyzeAExpr y tEnv vEnv) vEnv in
									let (x1,x2) = iVal2PairOfInt (analyzeAExpr x tEnv vEnv) vEnv in
									let (y1',y2') = iVal2PairOfInt (analyzeAExpr y' tEnv vEnv) vEnv in
									let (x1',x2') = iVal2PairOfInt (analyzeAExpr x' tEnv vEnv) vEnv in
									let yy1 = min y1 y1' in
									let xx1 = min x1 x1' in
									let yy2 = max y2 y2' in
									let xx2 = max x2 x2' 
									in (match v with 
										  AVal(b) -> 	if (((yy2-yy1+1)=(Array.length b)) && ((xx2-xx1+1)=(Array.length b.(0)))) then
															let _ = 
																for l=0 to (yy2-yy1) do
																	for k=0 to (xx2-xx1) do
																		a.(l+yy1-1).(k+xx1-1) <- cast b.(l).(k) (typeOf a.(l+yy1-1).(k+xx1-1))
																	done
																done  
													 		in (tEnv,setEnv vEnv id (AVal(a)))
														else
															raise (Error ("analyzeCmd","??? Matrix dimensions must agree."))
										| ArrayLines(_) -> raise (Error ("analyzeCmd","impossible in AssignArrayBlock"))
										| ArrayColumns(_) -> raise (Error ("analyzeCmd","impossible in AssignArrayBlock"))
					  					| _ -> let _ = 
														for l=(yy1-1) to (yy2-1) do
															for k=(xx1-1) to (xx2-1) do
																a.(l).(k) <- cast v (getArrayType (getEnvComplete id !typeEnv (Type_Array(I32(Int32.one),I32(Int32.one),typeOf v))))
															done
														done  
													 in (tEnv,setEnv vEnv id (AVal(a)))
										)
| 	Seq(c1,c2,Lab(line)) ->			let _ = currentLine := line in
									let (tEnv',vEnv') = analyzeCmd c1 tEnv vEnv 
									in analyzeCmd c2 tEnv' vEnv'
|	CondITE(e,c1,c2,Lab(line)) -> 	let _ = currentLine := line in
									let (trueVEnv,falseVEnv,tt) = analyzeBExpr e tEnv vEnv in
									let (trueTEnv',trueVEnv') = (match tt with
																	B(false) -> ([],[])
																|	BBottom ->	raise (Error ("analyzeCmd","bottom cond not expected"))
																| 	_ -> analyzeCmd c1 tEnv (insertEnv trueVEnv (copyEnv vEnv))
																)
									in
									let (falseTEnv',falseVEnv') = 	(match tt with
																		B(true) -> ([],[])
																	|	BBottom ->	raise (Error ("analyzeCmd","bottom cond not expected"))
																	| 	_ -> analyzeCmd c2 tEnv (insertEnv falseVEnv (copyEnv vEnv))
																	)
									in (tEnvJoin trueTEnv' falseTEnv',envJoin trueVEnv' falseVEnv')
|	While(e,c,Lab(line)) ->			let _ = currentLine := line in
									let (trueVEnv,falseVEnv,_) = analyzeBExpr e tEnv vEnv in
									let vEnv' = insertEnv trueVEnv vEnv in
									let (resTEnv,resVEnv) =	loopFP e c tEnv vEnv' (makeList [] (!unrollLevel) []) (makeList [] (!unrollLevel) []) 0 (getCondVarList e)
									in (resTEnv,envJoin resVEnv falseVEnv)
|	For(id,ae1,ae2,ae3,c,Lab(line)) ->	
									let _ = currentLine := line in
									if !forLoopWiden then
										let v1 = cast (analyzeAExpr ae1 tEnv vEnv) Type_Int_32 in
										let v2 = cast (analyzeAExpr ae2 tEnv vEnv) Type_Int_32 in
										let i = (match v1 with
													IVal(I32(a),_) -> 	Int32.to_int a
												|	_ ->				raise (Error("analyzeCmd","??? Incorrect lower bound in for loop."))
												)
										in
										let j = (match v2 with
													IVal(_,I32(a)) -> 	Int32.to_int a
												|	_ ->				raise (Error("analyzeCmd","??? Incorrect upper bound in for loop."))
												)
										in
										if (i<=j) then										
											let e = Lte(Id(id,Lab(line)),ae2,Lab(line)) in
											let inc = Assign(id,Plus(Id(id,Lab(line)),ae3,Lab(line)),Lab(line)) in
											let c' = Seq(c,inc,Lab(line)) in
											let cw = While(e,c',Lab(line)) in
											let ca = Assign(id,ae1,Lab(line)) in
											let c'' = Seq(ca,cw,Lab(line)) in 
												analyzeCmd c'' tEnv vEnv
										else
											let e = Gte(Id(id,Lab(line)),ae2,Lab(line)) in
											let ae3' = (match ae3 with 
															Cst(IVal(I32(a),I32(b)),lab) -> 	if ((Int32.compare a Int32.zero)>=0) then 
																								Cst(valNeg (IVal(I32(a),I32(b))),lab)
																							else ae3
														|   _ -> ae3)
											in
											let inc = Assign(id,Plus(Id(id,Lab(line)),ae3',Lab(line)),Lab(line)) in
											let c' = Seq(c,inc,Lab(line)) in
											let cw = While(e,c',Lab(line)) in
											let ca = Assign(id,ae1,Lab(line)) in
											let c'' = Seq(ca,cw,Lab(line)) in 
												analyzeCmd c'' tEnv vEnv											
									else
										let v1 = cast (analyzeAExpr ae1 tEnv vEnv) Type_Int_32 in
										let v2 = cast (analyzeAExpr ae2 tEnv vEnv) Type_Int_32 in
										let v3 = analyzeAExpr ae3 tEnv vEnv in
										let i = (match v1 with
													IVal(I32(a),_) -> 	Int32.to_int a
												|	IVal(I64(a),_) -> 	Int64.to_int a
												|	_ ->				raise (Error("analyzeCmd","??? Incorrect lower bound in for loop."))
												)
										in
										let j = (match v2 with
													IVal(_,I32(a)) -> 	Int32.to_int a
												|	IVal(_,I64(a)) -> 	Int64.to_int a
												|	_ ->				raise (Error("analyzeCmd","??? Incorrect upper bound in for loop."))
												)
										in
										let step = (match v3 with
														IVal(_,I32(a)) -> 	let s = Int32.to_int a in if (i<=j) then s else if s>0 then (-s) else s 
													|	IVal(_,I64(a)) -> 	let s = Int64.to_int a in if (i<=j) then s else if s>0 then (-s) else s
													|	_ ->				raise (Error("analyzeCmd","??? Incorrect step in for loop."))
													)
											in
										let tEnvRef = ref tEnv in
										let vEnvRef = ref vEnv in
										let k = ref i in
											if (i<=j) then
												(while ((!k)<=j) do 
													let (tEnv'',vEnv'') = analyzeCmd c (setEnv !tEnvRef id Type_Int_32) (setEnv !vEnvRef id (IVal(I32(Int32.of_int !k),I32(Int32.of_int !k)))) 
													in 	tEnvRef := tEnv'' ;
														vEnvRef := vEnv'' ;
														k := !k+step
													done ;
												 (!tEnvRef,!vEnvRef)
												)
											else
												(while ((!k)>=j) do
													let (tEnv'',vEnv'') = analyzeCmd c (setEnv !tEnvRef id Type_Int_32) (setEnv !vEnvRef id (IVal(I32(Int32.of_int !k),I32(Int32.of_int !k)))) 
													in 	tEnvRef := tEnv'' ;
														vEnvRef := vEnv'' ;
														k := !k+step
													done ;
												 (!tEnvRef,!vEnvRef)
												)


and loopFP e c tEnv vEnv unrollTEnvList unrollVEnvList nbIt condVarList =
	let (uTEnv,uVEnv) = (List.hd unrollTEnvList,List.hd unrollVEnvList) in
	if (nbIt<(!unrollLevel)) then
		let (itTEnv,itVEnv) = analyzeCmd c tEnv vEnv in
		let (trueEnv,falseEnv,tt) = analyzeBExpr e itTEnv itVEnv in
			(match tt with
				  B(t) -> 	if t then 
								let vEnv' = insertEnv trueEnv itVEnv in
								let unrollTEnvList' = (List.tl unrollTEnvList)@[itTEnv] in
								let unrollVEnvList' = (List.tl unrollVEnvList)@[vEnv'] in
									loopFP e c itTEnv vEnv' unrollTEnvList' unrollVEnvList' (nbIt+1) condVarList
							else 
								(itTEnv,insertEnv falseEnv itVEnv)
				| BTop -> 	let vEnv' = insertEnv trueEnv itVEnv in
							let unrollTEnvList' = (List.tl unrollTEnvList)@[itTEnv] in
							let unrollVEnvList' = (List.tl unrollVEnvList)@[vEnv'] in
								loopFP e c itTEnv vEnv' unrollTEnvList' unrollVEnvList' (nbIt+1) condVarList
			  	| BBottom -> raise (Error ("loopFP","??? Cannot evaluate loop condition."))
			)				
	else	
		let (t,uVEnvSseq) = sseqEnvNeglectErrorsList vEnv unrollVEnvList 
 		in	if t then 
				let vEnvJoinUVEnvSseq = envJoin vEnv uVEnvSseq in
	 			let (tEnv',vEnv') = analyzeCmd c uTEnv vEnvJoinUVEnvSseq in
				let (_,falseEnv,tt) = analyzeBExpr e tEnv' vEnv' 
				in (match tt with
					  	  B(t) -> if t then 
									raise (Error ("loopFP","??? Non-terminating loop."))
								else
									(tEnv',insertEnv falseEnv vEnvJoinUVEnvSseq)
						| BTop -> 	(tEnv',insertEnv falseEnv vEnvJoinUVEnvSseq)
						| BBottom -> raise (Error ("loopFP","??? Cannot evaluate loop condition."))
					)			
			else 
				if (nbIt>((!widenLevel)*(!unrollLevel))) then
					let (itTEnv,itVEnv) = analyzeCmd c tEnv vEnv in
					let vEnvWiden = envWiden vEnv itVEnv nbIt in
					let vEnvWiden' = envJoin vEnvWiden uVEnv in
					let (trueEnv,_,_) = analyzeBExpr e itTEnv vEnvWiden' in
					let vEnv' = envJoin (insertEnv trueEnv vEnvWiden') uVEnv in
					let unrollTEnvList' = (List.tl unrollTEnvList)@[itTEnv] in
					let unrollVEnvList' = (List.tl unrollVEnvList)@[vEnv'] 				
					in loopFP e c itTEnv vEnv' unrollTEnvList' unrollVEnvList' (nbIt+1) condVarList
				else
					let (itTEnv,itVEnv) = analyzeCmd c tEnv vEnv in
					let itVEnv' = envJoin itVEnv (*envJoin itVEnv vEnv*) uVEnv in
					let (trueEnv,_,_) = analyzeBExpr e itTEnv itVEnv' in
					let vEnv' = envJoin (insertEnv trueEnv itVEnv') uVEnv in
					let unrollTEnvList' = (List.tl unrollTEnvList)@[itTEnv] in
					let unrollVEnvList' = (List.tl unrollVEnvList)@[vEnv'] in
						loopFP e c itTEnv vEnv' unrollTEnvList' unrollVEnvList' (nbIt+1) condVarList


and valueVE ve tEnv vEnv = match ve with
	VECst(ArrayLines(id)) ->	let a = getEnv id vEnv 
								in (match a with
										AVal(av) -> IVal(I32(Int32.of_int (Array.length av)),I32(Int32.of_int (Array.length av)))
									|   _ ->		raise (Error ("valueVE","Impossible in ArrayLines case"))
								   ) 
|   VECst(ArrayColumns(id))	->	let a = getEnv id vEnv 
								in (match a with
										AVal(av) -> IVal(I32(Int32.of_int (Array.length av.(0))),I32(Int32.of_int (Array.length av.(0))))
									|   _ ->		raise (Error ("valueVE","Impossible in ArrayColumns case"))
							   	   ) 
|	VECst(c) -> 				c
|	VECast(t,ve1) -> 			let v = valueVE ve1 tEnv vEnv 
								in cast v t 
|	VEId(v,n) ->				let x = getEnvVal v vEnv 
								in (match x with
										FVal(x')	-> FVal(fePow x' n)
									|	IVal(x')	-> IVal(iPow x' n)
									|   XVal(x') ->    XVal(fiPow x' n) 
									|   AVal(x') ->    AVal(matrixPow x' n x')
									|	BVal(_) 	-> raise (Error ("ValueVE","Identifier "^v^"is boolean, should be numerical"))
									|	ValTop		-> ValTop
									|	ValBottom	-> ValBottom
									|   _ -> raise (Error ("valueVE","Impossible case in VEId")) 
									)
|	VEArrayElt(s,ve',ve'') -> 	let i = valueVE ve' tEnv vEnv in
								let j = valueVE ve'' tEnv vEnv in
								let (a,b) = iVal2PairOfInt i vEnv in
								let (c,d) = iVal2PairOfInt j vEnv in
									joinTabElts2D (getArray (getEnv s vEnv)) (a-1) (b-1) (c-1) (d-1) ValBottom 
|	VECreateMatrix(vel,matType) ->
								let il = List.map (fun x -> valueVE x tEnv vEnv) vel in
									evalVESpecialMatrix il vEnv matType 									
|	VEArrayBlock(s,ve1,ve2,ve3,ve4) -> 
								let i1 = valueVE ve1 tEnv vEnv in
								let i2 = valueVE ve2 tEnv vEnv in
								let i3 = valueVE ve3 tEnv vEnv in
								let i4 = valueVE ve4 tEnv vEnv in
								let (a1,b1) = iVal2PairOfInt i1 vEnv in
								let (a2,b2) = iVal2PairOfInt i2 vEnv in
								let (a3,b3) = iVal2PairOfInt i3 vEnv in
								let (a4,b4) = iVal2PairOfInt i4 vEnv in
								    if ((a1=b3)&&(a2=b4)) then
										let a = (getArray (getEnv s vEnv)) in
											a.(a1-1).(a2-1)
									else
										AVal(subMatrix (getArray (getEnv s vEnv)) (a1-1) (a2-1) (b3-1) (b4-1)) 
|	VEPlus(ve1,ve2) -> 			let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in
									 valPlus v1 v2
|	VEMinus(ve1,ve2) -> 		let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
									valMinus v1 v2
|	VETimes(ve1,ve2) -> 		let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
									valMult v1 v2
|	VEDotTimes(ve1,ve2) -> 		let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
									valDotMult v1 v2
|	VEDiv(ve1,ve2) -> 			let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
									valDiv v1 v2 
|	VEDotDiv(ve1,ve2) -> 		let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
									valDotDiv v1 v2
|	VEPower(ve1,ve2) -> 		let v1 = valueVE ve1 tEnv vEnv in
								let v2 = valueVE ve2 tEnv vEnv in 
								let e = iVal2PairOfInt v2 vEnv in
									valPower v1 e 
|	VEUMinus(ve1) -> 			let v1 = valueVE ve1 tEnv vEnv in
									valNeg v1 
|	VESqrt(ve1) -> 				let v1 = valueVE ve1 tEnv vEnv in
									valSqrt v1 
|	VEAbs(ve1) -> 				let v1 = valueVE ve1 tEnv vEnv in
									valAbs v1
|	VEMin(ve1) -> 				let v1 = valueVE ve1 tEnv vEnv in
									valArrayMin v1
|	VEMax(ve1) -> 				let v1 = valueVE ve1 tEnv vEnv in
									valArrayMax v1
|	VESize(ve1) -> 				let v1 = valueVE ve1 tEnv vEnv in
									valArraySize v1
|	VEFunCall(f,vel) ->         let il = List.map (fun x -> (Cst(valueVE x tEnv vEnv,Lab(!currentLine)))) vel 
								in (analyzeAExpr (FunCall(f,il,Lab(!currentLine))) tEnv vEnv)
| 	VEDeltaFun(f,vel1,vel2) ->  let vl1 = List.map (fun x -> (valueVE x tEnv vEnv)) vel1 in
								let vl2 = List.map (fun x -> (valueVE x tEnv vEnv)) vel2 in
								let vl = valListJoin vl1 vl2 [] in 
									computeDeltas f vl1 vl2 vl 0 ValBottom tEnv vEnv


and computeDeltas f vl1 vl2 vl n res tEnv vEnv = 
	if (n>=(List.length vl1)) then res
	else
		let res' = computeDelta f vl1 vl2 vl n tEnv vEnv in
		computeDeltas f vl1 vl2 vl (n+1) (valJoin res res') tEnv vEnv


and computeDelta f vl1 vl2 vl n tEnv vEnv =
	let (a,b,h) = computeDeltaParams (List.nth vl1 n) (List.nth vl2 n) in
	let res = ref ValBottom in
	let x = ref a in
	let _ =	while ((valLt (!x) b)=(B(true))) do
				let z = valJoin !x (valPlus (!x) h) in
				let vl' = setNth z n vl in
				let vl'' = List.map (fun x -> (Cst(x,Lab(!currentLine)))) vl' in
				let v = computeMaxHDelta (analyzeAExpr (FunCall(f,vl'',Lab(!currentLine))) tEnv vEnv) in
				let _ = if (valSseq !res v) then res := v else ()
				in x := valPlus !x h 
			done
	in
	let z = valJoin !x b in
	let vl' = setNth z n vl in
	let vl'' = List.map (fun x -> (Cst(x,Lab(!currentLine)))) vl' in
	let v = computeMaxHDelta (analyzeAExpr (FunCall(f,vl'',Lab(!currentLine))) tEnv vEnv) in
	let _ = if (valSseq !res v) then res := v else ()
	in !res
	

and evalFormal expr env k tEnv vEnv = match expr with
  Cst(ArrayLines(id),_) ->	let a = getEnv id vEnv 
							in (match a with
									AVal(av) -> [VECst(IVal(I32(Int32.of_int (Array.length av)),I32(Int32.of_int (Array.length av))))]
								|   _ ->		raise (Error ("evalFormal","Impossible in ArrayLines case"))
							   ) 
| Cst(ArrayColumns(id),_)	->	let a = getEnv id vEnv 
								in (match a with
										AVal(av) -> [VECst(IVal(I32(Int32.of_int (Array.length av.(0))),I32(Int32.of_int (Array.length av.(0)))))]
									|   _ ->		raise (Error ("evalFormal","Impossible in ArrayColumns case"))
							   	   ) 
| Cst(c,_) -> 						[VECst(c)]
| Cast(t,e,_) -> 					let veList = evalFormal e env k tEnv vEnv in
									List.map (fun x -> VECast(t,x)) veList
| Id(v,_) -> 						[VECst(getEnvVal v env); VEId(v,1)]  
| ArrayElt(a,e1,e2,_) -> 			let veList1 = evalFormal e1 env k tEnv vEnv in
									let	veList2 = evalFormal e2 env k tEnv vEnv in
										List.fold_left (fun y z -> ((List.map (fun x -> VEArrayElt(a,x,z)) veList1)@y)) [] veList2
| CreateMatrix(el,t,_) ->			[VECreateMatrix(List.map (fun x -> (VECst(evalExpr x tEnv vEnv))) el,t)]
| ArrayBlock(a,e1,e2,e3,e4,_) -> 	[VEArrayBlock(a,VECst(evalExpr e1 tEnv vEnv),VECst(evalExpr e2 tEnv vEnv),
													VECst(evalExpr e3 tEnv vEnv),VECst(evalExpr e4 tEnv vEnv))]				
| UndefinedCall(id,ael,_) ->		raise (Error ("evalFormal","undefinedCall not expected"))													
| Plus(e1,e2,_) -> 					let veList1  = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_PLUS k tEnv vEnv
| Minus(e1,e2,_) -> 				let veList1  = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_MINUS k tEnv vEnv
| Times(e1,e2,_) -> 			   	let veList1  = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
									let v = evalExpr e1 tEnv vEnv 
									in (match v with
										  AVal(_) -> combine veList1 veList2 OP_MAT_MULT k tEnv vEnv
										| _ -> combine veList1 veList2 OP_MULT k tEnv vEnv	
										)
| DotTimes(e1,e2,_) -> 			    let veList1  = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_DOT_MULT k tEnv vEnv
| Div(e1,e2,_) -> 		    		let veList1 = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_DIV k tEnv vEnv
| DotDiv(e1,e2,_) -> 			    let veList1  = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_DOT_DIV k tEnv vEnv
| Power(e1,e2,_) -> 		    	let veList1 = evalFormal e1 env k tEnv vEnv in
									let veList2 = evalFormal e2 env k tEnv vEnv in
										combine veList1 veList2 OP_POW k tEnv vEnv
| UMinus(Cst(c,_),_) -> 			[VEUMinus(VECst(c))]
| UMinus(e,_) -> 				    let veList = evalFormal e env k tEnv vEnv in
										(combine [VECst(IVal(!i32One))] veList OP_MULT k tEnv vEnv)@(List.map (fun ve -> (VEUMinus(ve))) veList) 
| Sqrt(e,_) ->						let veList = evalFormal e env k tEnv vEnv in
										List.map (fun ve -> (VESqrt(ve))) veList
| Abs(e,_) ->						let veList = evalFormal e env k tEnv vEnv in
										List.map (fun ve -> (VEAbs(ve))) veList
| Min(e,_) ->						let veList = evalFormal e env k tEnv vEnv in
										List.map (fun ve -> (VEMin(ve))) veList
| Max(e,_) ->						let veList = evalFormal e env k tEnv vEnv in
										List.map (fun ve -> (VEMax(ve))) veList
| Size(e,_) ->						let veList = evalFormal e env k tEnv vEnv in
										List.map (fun ve -> (VESize(ve))) veList
| FunCall(f,el,_) ->				[VEFunCall(f,List.map (fun x -> (VECst(evalExpr x tEnv vEnv))) el);VECst(analyzeAExpr expr tEnv vEnv)]						
| ArrayExpr(_) ->					raise (Error ("evalFormal","??? ArrayExpr not expected in evalFormal."))


and selectFloat veList vEnv tEnv bestVe bestVal bestWidth = match veList with
  [] -> (bestVe,valueVESlice bestVe tEnv vEnv !slicePrec)
| ve::ves -> let r = valueVE ve tEnv vEnv  
             in if (mpFloatLte (feWidthFloat r) bestWidth) then 
			      selectFloat ves vEnv tEnv ve r (feWidthFloat r)
				else
				  selectFloat ves vEnv tEnv bestVe bestVal bestWidth 
		

and selectError veList vEnv tEnv bestVe bestVal = match veList with
  [] -> (bestVe,valueVESlice bestVe tEnv vEnv !slicePrec)
| ve::ves -> let r = valueVE ve tEnv vEnv  
             in if (valErrLt r bestVal) then 
			      selectError ves vEnv tEnv ve r 
				else
				  selectError ves vEnv tEnv bestVe bestVal  


and selectComplex veList vEnv tEnv bestVe bestVal =
  let (bestVeF,bestValF) = selectFloat veList vEnv tEnv bestVe bestVal (feWidthFloat (valueVE bestVe tEnv vEnv)) in 								
  let (bestVeE,bestValE) = selectError veList vEnv tEnv bestVe bestVal in		
  if (valSseq bestValE bestValF) then
    (bestVeE,bestValE)
  else	
    match bestValF with
      FVal(FE(bestF,_,p)) ->
		let FE(_,bestE,_) = getFE bestValE 
        in (bestVeE,FVal(FE(bestF,bestE,p))) 
    | XVal(FX(bestX,ip,fp,_)) ->
		let FX(_,_,_,bestE) = getFX bestValE 
        in (bestVeE,XVal(FX(bestX,ip,fp,bestE))) 
    | _ -> (bestVe,bestVal) 


and combineAux ve veList op k tEnv vEnv = match veList with
  			[] -> []
		|   ve'::ves ->	let r = evalVE ve ve' op tEnv
						in insertList (removeK r k tEnv vEnv) (combineAux ve ves op k tEnv vEnv) 		  

			
and combine veList1 veList2 op k tEnv vEnv = match veList1 with
  			[] -> []
		  | ve::ves -> 	let vel1 = combineAux ve veList2 op k tEnv vEnv in
						let vel2 = combine ves veList2 op k tEnv vEnv in
							insertList vel1 vel2


(* slicing of expressions *)
and errorVESliceId ve tEnv vEnv k id globalRes leftRes =
	if (k>=(!slicePrec)) then
		let v = valueVE ve tEnv vEnv
		in valJoin leftRes v
	else
		let v = getEnv id vEnv in
		let (vLeft,vRight) = splitErr v in
		let resL = valueVE ve tEnv (setEnv vEnv id vLeft) in 
		let resR = valueVE ve tEnv (setEnv vEnv id vRight) in 					 
		let globalRes' = valJoin resL resR 
		in
			if (valSseq globalRes globalRes') then
				valJoin leftRes globalRes
			else
				let vL = errorVESliceId ve tEnv (setEnv vEnv id vLeft) (k+1) id globalRes' ValBottom 
				in errorVESliceId ve tEnv (setEnv vEnv id vRight) (k+1) id globalRes' (valJoin leftRes vL) 


and valueVESliceId ve tEnv vEnv k id globalRes leftRes =
	if (k>=(!slicePrec)) then 
		let v = valueVE ve tEnv vEnv
		in valJoin leftRes v
	else
		let v = getEnv id vEnv in
		let (vLeft,vRight) = splitVal v in
		let resL = valueVE ve tEnv (setEnv vEnv id vLeft) in 
		let resR = valueVE ve tEnv (setEnv vEnv id vRight) in 					 
		let globalRes' = valJoin resL resR 
		in
			if (valSseq globalRes globalRes') then
				valJoin leftRes globalRes
			else
				let vL = valueVESliceId ve tEnv (setEnv vEnv id vLeft) (k+1) id globalRes' ValBottom 
				in valueVESliceId ve tEnv (setEnv vEnv id vRight) (k+1) id globalRes' (valJoin leftRes vL) 

				
and valueVESliceIdById ve tEnv vEnv k idList bestV = match idList with
	[] -> 		let v = valueVE ve tEnv vEnv 
				in valMeet v bestV
|	id::ids	-> 	let t = (getEnv id tEnv) in
				(match t with
					Type_Int_32 
				| 	Type_Int_64
				|	Type_Float_32
				|	Type_Float_64
				|	Type_Float_128 
				|   Type_MP(_) 
				|   Type_Fixed(_) ->
						let v = valueVESliceId ve tEnv vEnv k id (valueVE ve tEnv vEnv) ValBottom in
						let v' = errorVESliceId ve tEnv vEnv k id (valueVE ve tEnv vEnv) ValBottom in
						let v'' = valMeet (valMeet v v') bestV in
							valueVESliceIdById ve tEnv vEnv k ids v'' 
				|   _ -> 
						let v = valueVE ve tEnv vEnv in
							valueVESliceIdById ve tEnv vEnv k ids v
				) 
					 
					
and valueVESlice ve tEnv vEnv k = 
	let veIdList = getVEVarList ve [] 
	in valueVESliceIdById ve tEnv vEnv k veIdList ValTop 			   


(* slicing of functions *)
and sliceErrorIdFunCall body tEnv vEnv k id outputs globalRes leftRes = 
	if (k>=(!slicePrec)) then 
		let (_,vEnv'') = analyzeCmd body tEnv vEnv in
		let v =	buildOutputVector outputs vEnv'' []
		in valJoin leftRes v
	else
		let v = getEnv id vEnv in
		let (vLeft,vRight) = splitErr v in
		let (_,resL) = analyzeCmd body tEnv (setEnv vEnv id vLeft) in 
		let (_,resR) = analyzeCmd body  tEnv (setEnv vEnv id vRight) in 
		let resL' =	buildOutputVector outputs resL [] in
		let resR' =	buildOutputVector outputs resR [] in				 
		let globalRes' = valJoin resL' resR' 
		in
			if (valSseq globalRes globalRes') then
				valJoin leftRes globalRes (* juste leftres ??*)
			else
				let vL = sliceErrorIdFunCall body tEnv (setEnv vEnv id vLeft) (k+1) id outputs globalRes' ValBottom 
				in sliceErrorIdFunCall body tEnv (setEnv vEnv id vRight) (k+1) id outputs globalRes' (valJoin leftRes vL) 


and sliceIdFunCall body tEnv vEnv k id outputs globalRes leftRes = 
	if (k>=(!slicePrec)) then 
		let (_,vEnv'') = analyzeCmd body tEnv vEnv in
		let v =	buildOutputVector outputs vEnv'' []
		in valJoin leftRes v
	else
		let v = getEnv id vEnv in
		let (vLeft,vRight) = splitVal v in
		let (_,resL) = analyzeCmd body tEnv (setEnv vEnv id vLeft) in 
		let (_,resR) = analyzeCmd body  tEnv (setEnv vEnv id vRight) in 
		let resL' =	buildOutputVector outputs resL [] in
		let resR' =	buildOutputVector outputs resR [] in		
		let globalRes' = valJoin resL' resR' 
		in
			if (valSseq globalRes globalRes') then
				valJoin leftRes globalRes (* juste leftres ??*)
			else
				let vL = sliceIdFunCall body tEnv (setEnv vEnv id vLeft) (k+1) id outputs globalRes' ValBottom 
				in sliceIdFunCall body tEnv (setEnv vEnv id vRight) (k+1) id outputs globalRes' (valJoin leftRes vL) 
	

and sliceIdsFunCall body argList outputs tEnv vEnv bestOV = match argList with 
	[] -> bestOV
|   id::ids -> 	let (_,vEnv') = analyzeCmd body tEnv vEnv in
				let ov = buildOutputVector outputs vEnv' [] in
				let outputVector = sliceIdFunCall body tEnv vEnv 0 id outputs ov ValBottom in
				let outputVector' = sliceErrorIdFunCall body tEnv vEnv 0 id outputs ov ValBottom in
				let outputVector'' = valMeet (valMeet outputVector outputVector') bestOV in
					sliceIdsFunCall body ids outputs tEnv vEnv outputVector'' 
				

and sliceFunCall id vel tEnv vEnv =
	let vl = List.map (fun x -> (analyzeAExpr x tEnv vEnv)) vel in
	let (tEnv',vEnv') = buildFunCallEnv id vl tEnv vEnv in
	let Fun(argList,outputs,body,_) = getEnv id !funEnv in
		sliceIdsFunCall body argList outputs tEnv' vEnv' ValTop


and flattenVE ve tEnv vEnv k = 
	if (k<=1) then
		VECst(valueVE ve tEnv vEnv)
	else
		(match ve with
		  VECst(_) -> ve
		| VECast(t,ve1) -> VECast(t,flattenVE ve1 tEnv vEnv (k-1))
		| VEArrayElt(_,_,_) -> ve
		| VEArrayBlock(_,_,_,_,_) -> ve
		| VECreateMatrix(vel,x) -> VECreateMatrix(List.map (fun y -> (flattenVE y tEnv vEnv (k-1))) vel,x) 
		| VEId(_) -> ve
		| VEPlus(ve1,ve2) -> VEPlus(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VEMinus(ve1,ve2) -> VEMinus(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VETimes(ve1,ve2) ->  VETimes(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VEDotTimes(ve1,ve2) ->  VEDotTimes(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VEDotDiv(ve1,ve2) ->  VEDotDiv(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VEDiv(ve1,ve2) ->  VEDiv(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1))
		| VEPower(ve1,ve2) -> VEPower(flattenVE ve1 tEnv vEnv (k-1),flattenVE ve2 tEnv vEnv (k-1)) 
		| VEUMinus(ve1) -> VEUMinus(flattenVE ve1 tEnv vEnv (k-1)) 
		| VESqrt(ve1) -> VESqrt(flattenVE ve1 tEnv vEnv (k-1))
		| VEAbs(ve1) -> VEAbs(flattenVE ve1 tEnv vEnv (k-1))
		| VEMin(ve1) -> VEMin(flattenVE ve1 tEnv vEnv (k-1))
		| VEMax(ve1) -> VEMax(flattenVE ve1 tEnv vEnv (k-1))
		| VESize(ve1) -> VESize(flattenVE ve1 tEnv vEnv (k-1))
		| VEFunCall(x,vel) -> VEFunCall(x,List.map (fun y -> (flattenVE y tEnv vEnv (k-1))) vel)
		| VEDeltaFun(x,vel1,vel2) -> VEDeltaFun(x,List.map (fun y -> (flattenVE y tEnv vEnv (k-1))) vel1,
											      List.map (fun y -> (flattenVE y tEnv vEnv (k-1))) vel2)
		) 
		

and removeK l k tEnv vEnv = List.map (fun x -> flattenVE x tEnv vEnv k) l 


and evalExpr expr tEnv vEnv = match expr with
  Cst(ArrayLines(id),_) ->	let a = getEnv id vEnv 
							in (match a with
									AVal(av) -> IVal(I32(Int32.of_int (Array.length av)),I32(Int32.of_int (Array.length av)))
								|   _ ->		raise (Error ("valueVE","Impossible in ArrayLines case"))
							   ) 
| Cst(ArrayColumns(id),_)	->	let a = getEnv id vEnv 
								in (match a with
										AVal(av) -> IVal(I32(Int32.of_int (Array.length av.(0))),I32(Int32.of_int (Array.length av.(0))))
									|   _ ->		raise (Error ("valueVE","Impossible in ArrayColumns case"))
							   	   ) 
| Cst(c,_) -> c
| Plus(e1,e2,_) -> valPlus (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| Minus(e1,e2,_) -> valMinus (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| Times(e1,e2,_) -> valMult (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| DotTimes(e1,e2,_) -> valDotMult (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| DotDiv(e1,e2,_) -> valDotDiv (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| Div(e1,e2,_) -> valDiv (evalExpr e1 tEnv vEnv) (evalExpr e2 tEnv vEnv)
| UMinus(e,_) -> valNeg (evalExpr e tEnv vEnv)
| Power(e1,e2,_) -> let e = iVal2PairOfInt (evalExpr e2 tEnv vEnv) vEnv
					in valPower (evalExpr e1 tEnv vEnv) e 
| Sqrt(e,_) -> valSqrt (evalExpr e tEnv vEnv)
| Abs(e,_) -> valAbs (evalExpr e tEnv vEnv)
| Min(e,_) -> valArrayMin (evalExpr e tEnv vEnv)
| Max(e,_) -> valArrayMax (evalExpr e tEnv vEnv)
| Size(e,_) -> valArraySize (evalExpr e tEnv vEnv)
| Cast(t,e,_) -> let v = evalExpr e tEnv vEnv in (cast v t)
| Id(v,_) -> getEnvVal v vEnv
| ArrayExpr(a,_) -> let x = (Array.length a.(0)) in
					let y = (Array.length a) in
					let b = Array.make_matrix y x ValBottom in
					let _ = 
						for i=0 to (y-1) do
							for j=0 to (x-1) do
								b.(i).(j) <- evalExpr a.(i).(j) tEnv vEnv
							done
						done
					in AVal(b)
| ArrayElt(s,e1,e2,_) -> 	let i1 = evalExpr e1 tEnv vEnv in
						 	let i2 = evalExpr e2 tEnv vEnv in  
						 	let (a1,b1) = iVal2PairOfInt i1 vEnv in
						 	let (a2,b2) = iVal2PairOfInt i2 vEnv in
						 		joinTabElts2D (getArray (getEnv s vEnv)) (a1-1) (b1-1) (a2-1) (b2-1) ValBottom 
| CreateMatrix(el,matType,_) ->	let il = List.map (fun x -> evalExpr x tEnv vEnv) el in
									evalVESpecialMatrix il vEnv matType 			
| ArrayBlock(s,e1,e2,e3,e4,_) ->    let i1 = evalExpr e1 tEnv vEnv in
									let i2 = evalExpr e2 tEnv vEnv in
									let i3 = evalExpr e3 tEnv vEnv in
									let i4 = evalExpr e4 tEnv vEnv in 
									let (a1,b1) = iVal2PairOfInt i1 vEnv in 
									let (a2,b2) = iVal2PairOfInt i2 vEnv in 
									let (a3,b3) = iVal2PairOfInt i3 vEnv in 
									let (a4,b4) = iVal2PairOfInt i4 vEnv in 
									    if ((a1=b3)&&(a2=b4)) then
											let a = (getArray (getEnv s vEnv)) in
												a.(a1-1).(a2-1)
										else
											AVal(subMatrix (getArray (getEnv s vEnv)) (a1-1) (a2-1) (b3-1) (b4-1))	
| FunCall(_) -> (analyzeAExpr expr tEnv vEnv)
| UndefinedCall(_) -> evalExpr (removeUndefinedCall expr tEnv vEnv) tEnv vEnv
;;

