(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file boolean.ml contains the domain and operations on boolean values
 *)

open Types ;;


let bJoin x y = match (x,y) with
	(B(true),B(true)) -> B(true)
 |  (B(false),B(false)) -> B(false)
 |  (B(true),B(false)) -> BTop
 |  (B(false),B(true)) -> BTop
 | 	(BTop,_) -> BTop
 |	(_,BTop) -> BTop
 |	(a,BBottom) -> a
 |	(BBottom,a) -> a ;;
 
 
 let bMeet x y = match (x,y) with
	(B(true),B(true)) -> B(true)
 |  (B(false),B(false)) -> B(false)
 |  (B(true),B(false)) -> BBottom
 |  (B(false),B(true)) -> BBottom
 | 	(BTop,a) -> a
 |	(a,BTop) -> a
 |	(a,BBottom) -> BBottom
 |	(BBottom,a) -> BBottom ;;
 
 
 let bSseq x y =  match (x,y) with
	(B(true),B(true)) -> true
 |  (B(false),B(false)) -> true
 |  (B(true),B(false)) -> false
 |  (B(false),B(true)) -> false
 |	(BTop,BTop) -> true
 |	(BTop,_) -> false  
 |	(_,BTop) -> true
 |	(BBottom,_) -> true
 |  (_,BBottom) -> false ;;

 
 let bAnd x y =  match (x,y) with
	(B(true),B(true)) -> B(true)
 |  (B(false),B(false)) -> B(false)
 |  (B(true),B(false)) -> B(false)
 |  (B(false),B(true)) -> B(false)
 |	(BBottom,_) -> BBottom
 |  (_,BBottom) -> BBottom 
 |	(BTop,B(false)) -> B(false)
 |	(BTop,_) -> BTop 
 |	(B(false),BTop) -> B(false)
 |	(_,BTop) -> BTop
;;


let bOr x y =  match (x,y) with
	(B(true),B(true)) -> B(true)
 |  (B(false),B(false)) -> B(false)
 |  (B(true),B(false)) -> B(true)
 |  (B(false),B(true)) -> B(true)
 |	(B(true),BTop) -> B(true)
 |	(BTop,B(true)) -> B(true)
 |	(BBottom,_) -> BBottom
 |  (_,BBottom) -> BBottom 
 |	(BTop,_) -> BTop 

 |	(_,BTop) -> BTop
;;


let bXor x y =  match (x,y) with
	(B(true),B(true)) -> B(false)
 |  (B(false),B(false)) -> B(false)
 |  (B(true),B(false)) -> B(true)
 |  (B(false),B(true)) -> B(true)
 |	(BBottom,_) -> BBottom
 |  (_,BBottom) -> BBottom 
 |	(BTop,_) -> BTop 
 |	(_,BTop) -> BTop
;;


let bNot x =  match x with
	B(true) -> B(false)
 |  B(false) -> B(true)
 |	BBottom -> BBottom
 |	BTop -> BTop
;;

