{
(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file lexer.mll contains the lexer
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Float ;;
open Fixed ;;
open Parser ;;


let create_hashtable size init =
    let tbl = Hashtbl.create size in
    List.iter (fun (key, data) -> Hashtbl.add tbl key data) init;
    tbl 


let keyword_table = 
    create_hashtable 50 [
      ("if", IF);
      ("elseif", ELSEIF);
      ("else", ELSE);
      ("while", WHILE);
      ("for", FOR);
      ("function", FUNCTION);
      ("inline", INLINE);
      ("end", END);
      ("int32", INT32);
      ("int", INT32);
      ("int64", INT64);
      ("long", INT64);
      ("float32", FLOAT32);
      ("float", FLOAT32);
      ("float64", FLOAT64);
      ("double", FLOAT64);
      ("float128", FLOAT128);
      ("quad", FLOAT128);
	  ("mp", FLOATMP);
	  ("fixed", FIXEDMP);
      ("set", SET);
      ("default", DEFAULT);
      ("formal", FORMAL);
      ("slice", SLICES);
      ("error", ERRPREC);
	  ("error_display",ERRPRECPRINT);
	  ("frac", FRAC);
      ("unroll", UNROLL);
      ("widen", WIDEN);
      ("bin", BINARY);
      ("binary", BINARY);
      ("dec", DECIMAL);
      ("absolute", ABSOLUTE);
      ("relative", RELATIVE);
      ("rel", RELATIVE);
      ("decimal", DECIMAL);
	  ("widen_for", WIDENFOR);
	  ("slice_fun", SLICEFUN);
	  ("debug", DEBUG);
      ("sqrt", SQRT);
      ("abs", ABS);
      ("eye", EYE);
      ("ones", ONES);
      ("zeros", ZEROS);
      ("rand", RAND);
      ("rand_errors", RANDERRORS);
      ("diag", DIAG);
      ("linspace", LINSPACE);
      ("logspace", LOGSPACE);
      ("size", SIZE);
      ("max", MAX);
      ("min", MIN);
      ("quit", EXIT);
      ("exit", EXIT);
      ("who", WHO);
      ("whos", WHOS);
      ("clear", CLEAR);
      ("cast", CAST);
      ("cd", CD);
      ("show", SHOW);
    ] 

}

rule token = parse
  ' '																	{ token lexbuf }
| '\t'                                                      	       	{ token lexbuf }
| "...\n"			                                            	    { token lexbuf }	
| "...\r"			                                            	    { token lexbuf }					
| '\n'                                								   	{ codeLine:= !codeLine+1; RET }
| '\r'																	{ codeLine:= !codeLine+1; RET }
| eof                                                              		{ EOF }
| (['0'-'9']*)'_'((['0'-'9']*))?										{ 
																			let s = Lexing.lexeme lexbuf 
																			in FIXED(fixedOfString s)
																		}
| (['0'-'9']*)'_'((['0'-'9']*))?('#'(['0'-'9']+))?						{ 
																			let s = Lexing.lexeme lexbuf 
																			in FIXED(fixedOfString s)
																		}
| (['0'-'9']*)('.')(['0'-'9']+)((('e'|'E')(('-'?)('+'?)(['0'-'9'])+))?)(('#'(['0'-'9']+))?)  	
																		{ 	let s = Lexing.lexeme lexbuf 
																			in if (String.contains s '#') then 
																			 		FLOAT(computeMP s)
																			   else
 																					FLOAT((M(FR.from_string_prec_base float64_mantissa GMP_RNDD 10 s),
														  								   M(FR.from_string_prec_base float64_mantissa GMP_RNDU 10 s)),Type_Float_64)
	 																	}				
| (['0'-'9']+)(('e'|'E')(('-'?)('+'?)(['0'-'9'])+))(('#'(['0'-'9']+))?)	  						
																		{
																		 	let s = Lexing.lexeme lexbuf 
																			in if (String.contains s '!') then 
																			 		FLOAT(computeMP s)
																			   else
	 																				FLOAT((M(FR.from_string_prec_base float64_mantissa GMP_RNDD 10 s),
														  								   M(FR.from_string_prec_base float64_mantissa GMP_RNDU 10 s)),Type_Float_64)
		 																}			
| (['0'-'9']+)('b'(('-'?)('+'?)(['0'-'9'])+))(('#'(['0'-'9']+))?)	  	{  
																			let s = Lexing.lexeme lexbuf 
																			in if (String.contains s '#') then 
																					let i = String.index s '#' in
																					let s' = String.sub s 0 i in 
																					let prec = int_of_string (String.sub s (i+1) ((String.length s)-i-1)) in
																					let (m,m',e) = extractMantissaAndExp s' in
																					let r =  mpFloatMult m (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) e prec) prec GMP_RNDD in
																					let r' =  mpFloatMult m' (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) e prec) prec GMP_RNDU
																					in FLOAT((r,r'),Type_MP(prec)) 
																			   else
																					let (m,m',e) = extractMantissaAndExp (Lexing.lexeme lexbuf) in
																		  			let r =  mpFloatMult m (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) e float64_mantissa) float64_mantissa GMP_RNDD in
																		  			let r' =  mpFloatMult m' (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) e float64_mantissa) float64_mantissa GMP_RNDU
																		  			in FLOAT((r,r'),Type_Float_64) 
																		}
| (['0'-'9']*)           	                                    		{ INT(I32(Int32.of_string(Lexing.lexeme lexbuf))) }
| '('                                                              		{ PARBEG }
| ')'                                                              		{ PAREND }
| '['                                                              		{ BRABEG }
| ']'                                                              		{ BRAEND }
| '*'		                                                            { MULT }
| ".*"																	{ DOTMULT }
| "./"																	{ DOTDIV }
| '+'                                                              		{ PLUS }
| '-'                                                              		{ MINUS }
| '/'                                                              		{ DIV }
| ';'                                                              		{ SC }
| ':'                                                              		{ COLON }
| ','                                                              		{ COMMA }
| "==" 																	{ EQ }
| "~=" 																	{ NEQ }
|">=" 																	{ GTE }
| ">" 																	{ GT }
| "<=" 																	{ LTE }
| '='                                                              		{ ASSIGN }
| '~'		                                                            { NOT }
| '&'		                                                            { AND }
| '<'		                                                            { LT }
| '|'		                                                            { OR }
| '^'		                                                            { HAT }
| '\''																	{ QUOTE }
| '%'(([^'\n''\r'])*) 					                                { token lexbuf }
| (['a'-'z']|['A'-'Z']|'_'|'\\'|'~'|"..")((['a'-'z']|['A'-'Z']|'_'|'/'|'\\'|'~'|".."|['0'-'9'])*) as word
																		{ try 
																			Hashtbl.find keyword_table word 
	  																	  with Not_found ->	ID(word)
																		} 
