(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file main.ml contains the main function and the interactive loop
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Float ;;
open Coercion ;;
open Print ;;
open Semiformal ;;
open Analyze ;;
 
let interactiveLoop () =  
	while (not (!theEnd)) do
		let _ = codeLine := 0 in
		let _ = warnMsgDivZero := false in
		let _ = warnMsgScalarProd := false in
		let _ = warnMsgMatProd := false in
		let _ = print_string "-] " ; flush stdout in 
		let lexbuf = Lexing.from_channel stdin in
		let (cmd,id) = 	try
							Parser.interaction Lexer.token lexbuf 
						with		Error(f,e) -> 	(if ((String.compare f "getEnv")=0) then
														print_string (e^"\n\n") 
													 else
													 	let d = if !debug then (" ("^f^")\n\n") else "\n\n"	
														in print_string (e^d)) ; (Nop,"NONE")
								|  	Parsing.Parse_error -> print_string ("??? Expression or statement is incorrect.\n\n") ; (Nop,"NONE")
								| 	Sys.Break ->	print_string "\n!!! Interrupted.\n" ; (Nop,"NONE") 
								|	e 			-> 	raise e 
		in
		let (tEnv',vEnv') = try
								let _ = funEnv := [] in 
								let (typeEnv',valEnv') = analyzeCmd cmd !typeEnv !valEnv in
								let _ = if (((String.compare id "NONE")!=0)&&(!verbose)) then 
											printValInteractive id typeEnv' valEnv'
								in (typeEnv',valEnv')
							with
								Error(f,e) 	->	let _ = if ((String.compare f "getEnv")=0) then
															print_string (e^"\n\n") 
														else	
															if ((!currentLine)>1) then
																let d = if !debug then (" ("^f^")\n\n") else "\n\n" 
																in print_string ("At line "^(string_of_int !currentLine)^": "^e^d)  
															else
																let d = if !debug then (" ("^f^")\n\n") else "\n\n" 
																in print_string (e^d) 
												in (!typeEnv,!valEnv)
							| 	Sys.Break ->	print_string "\n!!! Interrupted.\n" ; (!typeEnv,!valEnv)
							|	e 			->  raise e
		in
		let _ = typeEnv := tEnv' in
		let _ = valEnv := vEnv' in ()
	done ;;


(* Initialize random number generators *)
Random.self_init ();;

let rec initRand () = try
						RNG.randinit (RNG.GMP_RAND_ALG_LC (Random.int 18)) 
				  	  with _ -> initRand ();;

initRand() ;;

Sys.catch_break true ;;

print_string "\n  [ R A N G E . L A B ]\n";;

print_string "\n  Version 1.0\n" ;;
	
print_string "\n  September 2011\n" ;;

print_string "\n  RangeLab 1.0 Copyright (C) 2011 Matthieu Martel";;
print_string "\n  This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.";;
print_string "\n  This is free software, and you are welcome to redistribute it";;
print_string "\n  under certain conditions; type `show c' for details.\n\n";;

let rec loop () = try
	interactiveLoop() 
with 
	  Invalid_argument(e) -> (print_string ("??? Invalid argument: "^e^".\n\n") ; loop ())
	| _ -> (print_string "??? Unknown error.\n\n" ; loop ()) ;;

loop() ;;


print_string "\n" ;;

flush stdout ;;
