%{
(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file parser.ml contains the parser
 *)

open Types ;;
open Gmp ;;
open Prelude ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Coercion ;;
open Matrix ;;
open Arrays ;;
open Print ;;

%}

%type < Types.command * string > interaction
%start interaction
%token OR AND NOT PLUS MINUS MULT DOTMULT DOTDIV DIV HAT SQRT ABS
%token INT32 INT64 FLOAT32 FLOAT64 FLOAT128
%token EQ NEQ LT LTE GT GTE 
%token SC RET PARBEG PAREND QUOTE ACCBEG ACCEND
%token ASSIGN IF ELSEIF ELSE WHILE FOR FUNCTION INLINE END EOF
%token INT32 INT64 FLOAT32 FLOAT64 FLOAT128 FLOATMP FIXEDMP
%token BRABEG BRAEND COLON COMMA 
%token EXIT WHO WHOS CLEAR CAST CD
%token EYE ONES ZEROS RAND RANDERRORS DIAG EOF LINSPACE LOGSPACE MIN MAX SIZE
%token SET SLICES UNROLL WIDEN ERRPREC FORMAL DEFAULT BINARY DECIMAL ABSOLUTE RELATIVE WIDENFOR ERRPRECPRINT FRAC DEBUG SLICEFUN
%token SHOW
%token <string> ID 
%token <(Types.mpNumber * Types.mpNumber) * Types.typeDef> FLOAT
%token <(Types.fixedNumber * int * int) * Types.mpInterval> FIXED
%token <Types.intNumber> INT
%left OR AND NOT
%left PLUS MINUS
%left MULT DIV	
%right HAT

%%
interaction		: cmd RET										{ $1 }
				| WHO RET										{ verbose := true ; print_string ("\n"^(String.concat "" (List.map (fun (id,_) -> (if ( ((String.compare id "ans")=0) ||  ((String.compare id "pi")=0) ) then "" else (id^"  "))) (List.rev !valEnv)))^"\n\n") ; (Nop,"NONE") }
				| WHOS RET										{ verbose := true ; print_string ("\n"^(String.concat "" (List.map (fun (id,_) -> (if (((String.compare id "ans")=0)||((String.compare id "pi")=0)) then "" else (printWhos id))) (List.rev !valEnv)))^"\n\n") ; (Nop,"NONE") }
				| EXIT RET										{ theEnd := true ; (Nop,"NONE") }
				| SHOW ID RET									{ show $2 ;  (Nop,"NONE") }
				| CLEAR ID RET									{ valEnv := List.filter (fun (x,_) -> (String.compare $2 x)!=0) !valEnv ;
																  typeEnv := List.filter (fun (x,_) -> (String.compare $2 x)!=0) !typeEnv ; (Nop,"NONE") }
				| CLEAR	RET										{ typeEnv := (("pi",Type_Float_64)::[]) ;
																  valEnv := (("pi",FVal(pi))::[])  ; funEnv := [] ; (Nop,"NONE") }
				| aexpr	RET										{ verbose := true ; (Seq(VarDef("ans",snd $1,Lab(0)),Assign("ans",fst $1,Lab(0)),Lab(0)),"ans") }
				| aexpr SC RET 									{ verbose := false ; (Seq(VarDef("ans",snd $1,Lab(0)),Assign("ans",fst $1,Lab(0)),Lab(0)),"ans") }
				| bexpr	RET										{ verbose := true ; (Seq(VarDef("ans",Type_Bool,Lab(0)),AssignBool("ans",$1,Lab(0)),Lab(0)),"ans") }
				| bexpr SC RET 									{ verbose := false ; (Seq(VarDef("ans",Type_Bool,Lab(0)),AssignBool("ans",$1,Lab(0)),Lab(0)),"ans") }
				| RET											{ theEnd := false ; (Nop,"NONE") }
				| RET interaction                               { $2 }
				| FUNCTION ID ASSIGN ID PARBEG argList PAREND cmdBlock EOF
																{ funEnv := setEnv !funEnv $4 (Fun($6,$2::[],list2Seq $8,Lab(!codeLine))) ; (Nop,"FUNCTION") }
				| FUNCTION BRABEG argList BRAEND ASSIGN ID PARBEG argList PAREND cmdBlock EOF
																{ funEnv := setEnv !funEnv $6 (Fun($8,$3,list2Seq $10,Lab(!codeLine))) ; (Nop,"FUNCTION") }

				| set											{ (Nop,"NONE") }
;
cmdLine			: ID ASSIGN aexpr								{ (Seq(VarDef($1,snd $3,Lab(!codeLine)),Assign($1,fst $3,Lab(!codeLine)),Lab(!codeLine)),$1) }
				| ID PARBEG aexprList PAREND ASSIGN aexpr		{ let l = List.length $3 in
																	if (l=1) then
																		let (m,n) = try (getArrayLength (getEnv $1 !typeEnv),
																	  	 			 	 getArrayLineLength (getEnv $1 !typeEnv))
																					with
																						Error(_) -> (I32(Int32.of_int 2),I32(Int32.one))
																					|	e -> raise e
																	  	in if ((m=(I32(Int32.one)))||(m=(I64(Int64.one)))) then				
																				(AssignArrayElt($1,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),fst (List.hd $3),fst $6,Lab(!codeLine)),$1) 
																		   else
																				if ((n=(I32(Int32.one)))||(n=(I64(Int64.one)))) then
																					(AssignArrayElt($1,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),fst (List.hd $3),fst $6,Lab(!codeLine)),$1) 
																				else raise (Error ("Parser",("??? In left hand side of assignment, dimensions must agree."^(printIVal (m,m))^"hh"^(printIVal (n,n)))))
																	else
																		if (l=2) then
																			(AssignArrayElt($1,fst (List.hd $3),fst (List.hd (List.tl $3)),fst $6,Lab(!codeLine)),$1)
																		else
																		 	raise (Error ("parser","??? In left hand side of assignment, dimensions must agree."))
																}	
				| assignArrayBlock								{ $1 }
				| IF bexpr sep cmdBlock elseif					{ (CondITE($2,list2Seq $4,$5,Lab(!codeLine)),"NONE") }			
				| WHILE bexpr sep cmdBlock END	 				{ (While($2,list2Seq $4,Lab(!codeLine)),"NONE") }
				| FOR ID ASSIGN aexprScalar COLON aexprScalar sep cmdBlock END 
																{ (For($2,fst $4,fst $6,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),list2Seq $8,Lab(!codeLine)),"NONE") }
				| FOR ID ASSIGN aexprScalar COLON aexprScalar sep cmdBlock COMMA END 
																{ (For($2,fst $4,fst $6,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),list2Seq $8,Lab(!codeLine)),"NONE") }
				| FOR ID ASSIGN aexprScalar COLON aexprScalar COLON aexprScalar sep cmdBlock END 
																{ (For($2,fst $4,fst $8,fst $6,list2Seq $10,Lab(!codeLine)),"NONE") }										
				| FOR ID ASSIGN aexprScalar COLON aexprScalar COLON aexprScalar sep cmdBlock COMMA END 
																{ (For($2,fst $4,fst $8,fst $6,list2Seq $10,Lab(!codeLine)),"NONE") }
;
sep				: RET											{ }
				| COMMA											{ }
				;
elseif			: END										{ Nop }
				| ELSE cmdBlock END							{ list2Seq $2 }
				| ELSEIF bexpr sep cmdBlock elseif			{ CondITE($2,list2Seq $4,$5,Lab(!codeLine)) }
				;
argList			: ID											{ $1::[] }
				| ID COMMA argList								{ $1::$3 }
;
cmd 			: cmdLine 	 									{ verbose := true ; $1 }
				| cmdLine SC 	 								{ verbose := false ; $1 }
				| cmdLine SC cmd								{ (Seq(fst $1,fst $3,Lab(!codeLine)),snd $3) }
;
cmdBlock		: RET											{ [] } 
				| cmd											{ (fst $1)::[] }
				| cmd COMMA										{ (fst $1)::[] }
				| cmdBlock cmdBlock								{ $1@$2 }
;
aexpr			: array											{ $1 }
				| aexprScalar									{ $1 }
;
aexprScalar     : number			    	      		  	    { (Cst(fst $1,Lab(!codeLine)),snd $1) }
				| typeDef PARBEG aexprScalar PAREND 			{ (Cast($1,fst $3,Lab(!codeLine)),$1) }
				| CAST PARBEG aexpr COMMA QUOTE typeDef QUOTE PAREND
																{ (Cast($6,fst $3,Lab(!codeLine)),$6) }
   				| ID                                			{ (Id($1,Lab(!codeLine)),getEnvComplete $1 !typeEnv UndefinedType) }
				| ID PARBEG aexprList PAREND					{ (UndefinedCall($1,List.map fst $3,Lab(!codeLine)),UndefinedType) }
				| subArray										{ $1 }
				| aexpr PLUS aexpr            				   	{ (Plus(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| aexpr MINUS aexpr  	       				   	{ (Minus(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| aexpr MULT aexpr  	       				   	{ (Times(fst $1,fst $3,Lab(!codeLine)),maxTypeMatrixMult (snd $1) (snd $3)) }
				| aexpr DOTMULT aexpr  	       				   	{ (DotTimes(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| aexpr DOTDIV aexpr  	       				   	{ (DotDiv(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| aexpr DIV aexpr  	           				   	{ (Div(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| aexpr HAT aexpr								{ (Power(fst $1,fst $3,Lab(!codeLine)),maxType (snd $1) (snd $3)) }
				| SQRT PARBEG aexpr PAREND						{ (Sqrt(fst $3,Lab(!codeLine)),snd $3) }
				| ABS PARBEG aexpr PAREND						{ (Abs(fst $3,Lab(!codeLine)),snd $3) }
				| MINUS aexpr  				                    { (UMinus(fst $2,Lab(!codeLine)),snd $2) }
				| MIN PARBEG aexpr PAREND						{ (Min(fst $3,Lab(!codeLine)),snd $3) }
				| MAX PARBEG aexpr PAREND						{ (Max(fst $3,Lab(!codeLine)),snd $3) }
				| SIZE PARBEG aexpr PAREND						{ (Size(fst $3,Lab(!codeLine)),snd $3) }				
				| PARBEG aexpr PAREND	                        { $2 }
;
aexprList		: aexprScalar									{ $1::[] }
				| aexprScalar COMMA aexprList					{ $1::$3 }
;
bexpr			: aexpr EQ aexpr								{ Eq(fst $1,fst $3,Lab(!codeLine)) }
				| aexpr NEQ aexpr								{ Neq(fst $1,fst $3,Lab(!codeLine)) }
				| aexpr LT aexpr								{ Lt(fst $1,fst $3,Lab(!codeLine)) }
				| aexpr LTE aexpr								{ Lte(fst $1,fst $3,Lab(!codeLine)) }
				| aexpr GT aexpr								{ Gt(fst $1,fst $3,Lab(!codeLine)) }
				| aexpr GTE aexpr								{ Gte(fst $1,fst $3,Lab(!codeLine)) }
				| bexpr AND bexpr								{ And($1,$3,Lab(!codeLine)) }
				| bexpr OR bexpr								{ Or($1,$3,Lab(!codeLine)) }
				| bexpr HAT bexpr								{ Xor($1,$3,Lab(!codeLine)) }
				| NOT bexpr										{ Not($2,Lab(!codeLine)) }				
				| PARBEG bexpr PAREND	                        { $2 }
;
typeDef			: INT32											{ Type_Int_32 }
				| INT64											{ Type_Int_64 }
				| FLOAT32										{ Type_Float_32 }
				| FLOAT64										{ Type_Float_64 }
				| FLOAT128										{ Type_Float_128 }
				| FLOATMP PARBEG INT PAREND 					{ Type_MP(getInt $3) }
				| FIXEDMP PARBEG INT COMMA INT PAREND           { Type_Fixed(getInt $3,getInt $5) }
				| typeDef BRABEG INT COMMA INT BRAEND			{ Type_Array($3,$5,$1) }
;
signedInt		: INT											{ $1 }
				| MINUS INT 									{ integerMult (I32(Int32.of_int (-1))) $2 }			
				;		
signedFixed		: FIXED											{ $1 }
				| MINUS FIXED									{ let ((x,ip,fp),e) = $2 in ((fixedNeg x,ip,fp),e)  }
				;
signedFloat		: FLOAT											{ (match (snd $1) with
																		Type_Float_64 -> (FVal(FE(fst $1,initError (fst $1),float64_mantissa)),Type_Float_64)
																   	|	Type_MP(p) -> (FVal(FE(fst $1,initError (fst $1),p)),snd $1)
																	|	_ -> raise (Error("Parser","Impossible case for floats"))
																   )
															 	}
				| MINUS FLOAT									{ (match (snd $2) with
																		Type_Float_64 -> let mf = mpfiNeg (fst $2) float64_mantissa OUTSIDE 
																		in (FVal(FE(mf,initError (fst $2),float64_mantissa)),Type_Float_64)
																   	|	Type_MP(p) -> let mf = mpfiNeg (fst $2) p OUTSIDE 
																		in (FVal(FE(mf,initError (fst $2),p)),Type_MP(p))
																	|	_ -> raise (Error("Parser","Impossible case for floats"))
																	)
															 	}
				;
singletonNumber : signedInt										{ (IVal($1,$1),Type_Int_32) }
				| signedFloat									{ $1 }
				| signedFixed   								{ let ((x,ip,fp),e) = $1 in (XVal(FX((x,x),ip,fp,e)),Type_Fixed(ip,fp)) }
				;
number			: INT											{ (IVal($1,$1),Type_Int_32) }
				| FLOAT											{ (match (snd $1) with
																		Type_Float_64 -> (FVal(FE(fst $1,initError (fst $1),float64_mantissa)),Type_Float_64)
																   	|	Type_MP(p) -> (FVal(FE(fst $1,initError (fst $1),p)),snd $1)
																	|	_ -> raise (Error("Parser","Impossible case for floats"))
																	)
															 	}
				| FIXED											{ let ((x,ip,fp),e) = $1 in (XVal(FX((x,x),ip,fp,e)),Type_Fixed(ip,fp)) }
				| BRABEG singletonNumber COMMA singletonNumber BRAEND
																{
																	let t = maxType (snd $2) (snd $4) in
																	let v1 = cast (fst $2) t in
																	let v2 = cast (fst $4) t in (valJoin v1 v2,t)
																}
				| BRABEG singletonNumber COMMA singletonNumber COMMA singletonNumber COMMA singletonNumber BRAEND
																{
																	let tMax = maxType (snd $2) (snd $4) in
																	let t = (match tMax with
																				Type_Int_32 -> Type_Float_64
																			| 	Type_Int_64 -> Type_Float_64
																			|	_ -> tMax
																			)
																	in match t with
																			Type_Int_32 -> raise (Error ("parser","??? Cannot attach errors to integers."))
																		|   Type_Int_64 -> raise (Error ("parser","??? Cannot attach errors to integers."))
																		|   Type_Bool -> raise (Error ("parser","impossible case"))
																		|   Type_Array(_) -> raise (Error ("parser","impossible case"))																		
																		|   Type_Fun(_) -> raise (Error ("parser","impossible case"))																		
																		|   UndefinedType -> raise (Error ("parser","impossible case"))
																		|   _ -> 	let v1 = cast (fst $2) t in
																					let v2 = cast (fst $4) t in 
																					let v3 = fValFst (cast (fst $6) (Type_MP(!errPrec))) in 
																					let v4 = fValSnd (cast (fst $8) (Type_MP(!errPrec))) in 
																					let v = valJoin v1 v2
																					in (match v with
																						  FVal(FE((f1,f2),_,p)) -> (FVal(FE((f1,f2),(v3,v4),p)),t)
																						| XVal(FX((x,y),ip,fp,_)) -> (XVal(FX((x,y),ip,fp,(v3,v4))),t)
																						| _ -> raise (Error ("parser","impossible case"))
																						)
																}
				;
array			: BRABEG arrayLines BRAEND 						{ let a = processArray (fst $2) in 
																  (ArrayExpr(a,Lab(!codeLine)),
																  Type_Array(I32(Int32.of_int (List.length (fst $2))),I32(Int32.of_int (List.length (List.hd (fst $2)))),snd $2)) }
				| BRABEG arrayLines BRAEND QUOTE				{ let a = processArray (fst $2) in 
																  let a' = ArrayExpr(a,Lab(!codeLine)) in
																  (CreateMatrix(a'::[],Matrix_Transpose,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
;
arrayLines		: arrayLine										{ (( (fst $1))::[],snd $1) }
				| arrayLine SC arrayLines						{ (( (fst $1))::(fst $3),maxType (snd $1) (snd $3)) }
;
arrayLine		: aexprScalar 									{ (fst $1::[],snd $1) }
				| aexprScalar arrayLine							{ ((fst $1)::(fst $2),maxType (snd $1) (snd $2))  }
;
subArray		: ID PARBEG aexprScalar COMMA COLON PAREND		{ let d = ArrayColumns($1)  
																  in (ArrayBlock($1,fst $3,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																					fst $3,Cst(d,Lab(!codeLine)),Lab(!codeLine)),
																	  UndefinedType) 
																}
				| ID PARBEG aexprScalar COLON aexprScalar PAREND			
																{ let d = ArrayColumns($1)  
																  in (ArrayBlock($1,fst $3,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																					fst $5,Cst(d,Lab(!codeLine)),Lab(!codeLine)),
																	  UndefinedType) 
																}
				| ID PARBEG aexprScalar COLON aexprScalar COMMA COLON PAREND
																{ let d = ArrayColumns($1)  
																  in (ArrayBlock($1,fst $3,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																					fst $5,Cst(d,Lab(!codeLine)),Lab(!codeLine)),
																	  UndefinedType) 
																}
				| ID PARBEG COLON COMMA aexprScalar COLON aexprScalar PAREND
																{ let d = ArrayLines($1) 
																  in (ArrayBlock($1,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),fst $5,
																					Cst(d,Lab(!codeLine)),fst $7,Lab(!codeLine)),
																	  UndefinedType) 
																}																	
				| ID PARBEG COLON COMMA aexprScalar PAREND 		{ let d = ArrayLines($1)
																  in (ArrayBlock($1,Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),fst $5,
																					Cst(d,Lab(!codeLine)),fst $5,Lab(!codeLine)),
																	  UndefinedType) 
																}																	
				| ID PARBEG aexprScalar COLON aexprScalar COMMA aexprScalar COLON aexprScalar PAREND 	
																{ 
																  (ArrayBlock($1,fst $3,fst $7,fst $5,fst $9,Lab(!codeLine)),
																   UndefinedType) 
																}
				| ID PARBEG aexprScalar COMMA aexprScalar COLON aexprScalar PAREND 	
																{ 
																  (ArrayBlock($1,fst $3,fst $5,fst $3,fst $7,Lab(!codeLine)),
																   UndefinedType)
																}
				| ID PARBEG aexprScalar COLON aexprScalar COMMA aexprScalar PAREND 	
																{ 
																  (ArrayBlock($1,fst $3,fst $7,fst $5,fst $7,Lab(!codeLine)),
																   UndefinedType)
																}
				| EYE PARBEG aexprScalar PAREND					{ (CreateMatrix((fst $3)::[],Matrix_Eye,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| ONES PARBEG aexprScalar COMMA aexprScalar PAREND	
																{ (CreateMatrix((fst $3)::(fst $5)::[],Matrix_Ones,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| ONES PARBEG aexprScalar PAREND	
																{ (CreateMatrix((fst $3)::(fst $3)::[],Matrix_Ones,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| ZEROS PARBEG aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::[],Matrix_Zeros,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| ZEROS PARBEG aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $3)::[],Matrix_Zeros,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }														
				| RAND PARBEG aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::[],Matrix_Rand,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }		
				| RAND PARBEG aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $3)::[],Matrix_Rand,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }		
				| RANDERRORS PARBEG aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::[],Matrix_Rand_Errors,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| RANDERRORS PARBEG aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $3)::[],Matrix_Rand_Errors,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| DIAG PARBEG aexprScalar PAREND				{ (CreateMatrix((fst $3)::(Cst(IVal(!i32One),Lab(!codeLine)))::[],Matrix_Diag,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| DIAG PARBEG aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::[],Matrix_Diag,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }				
				| aexprScalar QUOTE								{ (CreateMatrix((fst $1)::[],Matrix_Transpose,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }			
				| BRABEG aexprScalar COLON aexprScalar BRAEND 	{ (CreateMatrix((fst $2)::(fst $4)::(Plus(Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(0)),Minus(fst $4,fst $2,Lab(0)),Lab(0)))::[],Matrix_Linspace,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| LINSPACE PARBEG aexprScalar COMMA aexprScalar PAREND  
																{ (CreateMatrix((fst $3)::(fst $5)::(Cst(IVal(I32(Int32.of_int 100),I32(Int32.of_int 100)),Lab(0)))::[],Matrix_Linspace,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }	
				| LINSPACE PARBEG aexprScalar COMMA aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::(fst $7)::[],Matrix_Linspace,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }																	
				| LOGSPACE PARBEG aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::(Cst(IVal(I32(Int32.of_int 50),I32(Int32.of_int 50)),Lab(0)))::[],Matrix_Logspace,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }	
				| LOGSPACE PARBEG aexprScalar COMMA aexprScalar COMMA aexprScalar PAREND
																{ (CreateMatrix((fst $3)::(fst $5)::(fst $7)::[],Matrix_Logspace,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) }
				| BRABEG aexprScalar COLON aexprScalar COLON aexprScalar BRAEND
																{ (CreateMatrix((fst $2)::(fst $4)::(fst $6)::[],Matrix_Colon_Step,Lab(!codeLine)),Type_Array(IBottom,IBottom,Type_Float_64)) } 
;
assignArrayBlock: ID PARBEG COLON COMMA aexprScalar PAREND ASSIGN aexpr		
																{ let d = ArrayLines($1) 
												  				  in (AssignArrayBlock($1,
																 	   				   Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																	   				   fst $5,
																	   				   Cst(d,Lab(!codeLine)),
																	    			   fst $5,
																	   				   fst $8,Lab(!codeLine)),$1)
																}
				| ID PARBEG aexprScalar COMMA COLON PAREND ASSIGN aexpr		
																{ let d = ArrayColumns($1) 
												  				  in (AssignArrayBlock($1,
																 	   				   fst $3,
																					   Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																	   				   fst $3,
																	   				   Cst(d,Lab(!codeLine)),
																	   				   fst $8,Lab(!codeLine)),$1)
																}
				| ID PARBEG aexprScalar COLON aexprScalar COMMA aexprScalar PAREND ASSIGN aexpr		
																{ (AssignArrayBlock($1,
												 	   			   				    fst $3, 
																				    fst $7,
															   				   	    fst $5,
														    			   	   		fst $7,
																   				    fst $10,Lab(!codeLine)),$1)
																}			
				| ID PARBEG aexprScalar COMMA aexprScalar COLON aexprScalar PAREND ASSIGN aexpr		
																{ (AssignArrayBlock($1,
								 	   			   				    				fst $3, 
																    				fst $5,
											   				   	    				fst $3,
										    			   	   						fst $7,
												   				    				fst $10,Lab(!codeLine)),$1)
																}			
				| ID PARBEG aexprScalar COLON aexprScalar COMMA aexprScalar COLON aexprScalar PAREND ASSIGN aexpr		
																{ (AssignArrayBlock($1,
												   				    				fst $3, 
																    				fst $7,
																    				fst $5,
									    			   	   							fst $9,
												   				    				fst $12,Lab(!codeLine)),$1)
																}
				| ID PARBEG aexprScalar COLON aexprScalar PAREND ASSIGN aexpr		
																{ let d = ArrayColumns($1) 
											  				  	  in (AssignArrayBlock($1,
															 	   				   	   fst $3,
																					   Cst(IVal(I32(Int32.one),I32(Int32.one)),Lab(!codeLine)),
																	 				   fst $5,
																	   				   Cst(d,Lab(!codeLine)),
																   				   	   fst $8,Lab(!codeLine)),$1)
																}
;
set				: SET SLICES INT RET							{ slicePrec := getInt $3 }
				| SET WIDEN INT	RET								{ widenLevel := (getInt $3) }
				| SET UNROLL INT RET							{ unrollLevel := getInt $3 }
				| SET ERRPREC INT RET							{ errPrec := getInt $3 }
				| SET ERRPRECPRINT INT RET						{ errPrecPrint := getInt $3 }
				| SET FRAC INT RET								{ fixed_default_frac_size := getInt $3 }
				| SET FORMAL INT RET							{ formalExprMaxHeight := getInt $3 }
				| SET BINARY									{ printBinary := true }
				| SET DECIMAL									{ printBinary := false }
				| SET ABSOLUTE									{ printAbsolute := true }
				| SET ABS										{ printAbsolute := true }
				| SET RELATIVE									{ printAbsolute := false }
				| SET DEBUG										{ debug := not (!debug) }
				| SET SLICEFUN									{ noFunCallYet := not (!noFunCallYet) }
				| SET WIDENFOR									{ forLoopWiden := not (!forLoopWiden) }
				| SET RET										{ printEnvVars () }
				| SET DEFAULT RET								{ setDefault () }
				| CD PARBEG QUOTE path QUOTE PAREND RET			{ changeDir $4 }
;
path			: ID DIV path									{ ($1)^"/"^$3 }
				| DIV path										{ "/"^$2 }
				| ID											{ $1 }
;;

%%
