(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file widenings.ml contains the widening operators for the different types of values
 *)

open Types ;;
open Prelude ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Gmp ;;
open Coercion ;;

(*****************************************************************************)
(*                        Widening on ground types                           *)
(*****************************************************************************)


let rec i32NextPowOfTwoAux i n = 
	if n>30 then IPInfty
	else
		let w = integerPow (I32(Int32.of_int 2)) n
		in 	if (integerLt w i) then
		  		i32NextPowOfTwoAux i (n+1)
			else w ;;
			
		
let i32NextPowOfTwo i nbIt = 
	if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then IPInfty 
	else i32NextPowOfTwoAux i (int_of_float (log (float_of_int (getInt i))/.(log 2.0))) ;;


let i32NextPowOfTwoNeg i nbIt = 
	if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then IMInfty 
	else
		let i' = integerNeg i in
		let w = i32NextPowOfTwoAux i' (int_of_float (log (float_of_int (getInt i'))/.(log 2.0))) 
		in integerNeg w ;; 
		

let rec i64NextPowOfTwoAux i n = 
	if n>62 then IPInfty
	else
		let w = integerPow (I64(Int64.of_int 2)) n
		in 	if (integerLt w i) then
		  		i64NextPowOfTwoAux i (n+1)
			else w ;;
			
		
let i64NextPowOfTwo i nbIt = 
	if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then IPInfty 
	else i64NextPowOfTwoAux i (int_of_float (log (float_of_int (getInt i))/.(log 2.0))) ;;


let i64NextPowOfTwoNeg i nbIt = 
	if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then IMInfty 
	else
		let i' = integerNeg i in
		let w = i64NextPowOfTwoAux i' (int_of_float (log (float_of_int (getInt i'))/.(log 2.0))) 
		in integerNeg w ;; 
		
		
let rec integerWidenRight a b nbIt = match (a,b) with
	(IBottom,_) -> 			b
|	(I32(_),I32(_))	->		if ((integerLte a (I32(Int32.zero)))&&(integerEq b (I32(Int32.zero)))) then (I32(Int32.zero)) 
							else  
								if (integerLte b a) then a
								else 
									if (integerLt b (I32(Int32.zero))) then (I32(Int32.zero)) else (i32NextPowOfTwo b nbIt)
|	(I64(_),I64(_)) ->		if ((integerLte a (I64(Int64.zero)))&&(integerEq b (I64(Int64.zero)))) then (I64(Int64.zero)) 
							else  
								if (integerLte b a) then a
								else 
									if (integerLt b (I64(Int64.zero))) then (I64(Int64.zero)) else (i64NextPowOfTwo b nbIt)
|	(I32(_),I64(_)) ->		integerWidenRight (castInt2Int64 a) b nbIt
|	(I64(_),I32(_)) ->		integerWidenRight a (castInt2Int64 b) nbIt
|	(IMInfty,I32(_))->		if (integerEq b (I32(Int32.zero))) then (I32(Int32.zero)) 
							else  
								if (integerLt b (I32(Int32.zero))) then (I32(Int32.zero)) else (i32NextPowOfTwo b nbIt)
|	(IMInfty,I64(_))->		if (integerEq b (I64(Int64.zero))) then (I64(Int64.zero)) 
							else  
								if (integerLt b (I64(Int64.zero))) then (I64(Int64.zero)) else (i64NextPowOfTwo b nbIt)
|	(IMInfty,_)		-> 		b
|	(_,IMInfty)		->		a
|	(IPInfty,_)		->		IPInfty
|	(_,IPInfty)		->		IPInfty				
|	(_,IBottom) ->			a			
;;


let rec integerWidenLeft a b nbIt = match (a,b) with
	(_,IBottom) -> 			a
|	(I32(_),I32(_))	->		if ((integerGte a (I32(Int32.zero)))&&(integerEq b (I32(Int32.zero)))) then (I32(Int32.zero)) 
							else
								if (integerGte b a) then a 
								else 
									if (integerGt b (I32(Int32.zero))) then (I32(Int32.zero)) else (i32NextPowOfTwoNeg b nbIt) 
|	(I64(_),I64(_)) ->		if ((integerGte a (I64(Int64.zero)))&&(integerEq b (I64(Int64.zero)))) then (I64(Int64.zero)) 
							else
								if (integerGte b a) then a 
								else 
									if (integerGt b (I64(Int64.zero))) then (I64(Int64.zero)) else (i64NextPowOfTwoNeg b nbIt)
|	(I32(_),I64(_)) ->		integerWidenLeft (castInt2Int64 a) b nbIt
|	(I64(_),I32(_)) ->		integerWidenLeft a (castInt2Int64 b) nbIt
|	(IMInfty,I32(_))->		if (integerEq b (I32(Int32.zero))) then (I32(Int32.zero)) 
							else
								if (integerGt b (I32(Int32.zero))) then (I32(Int32.zero)) else (i32NextPowOfTwoNeg b nbIt)
|	(IMInfty,I64(_))->		if (integerEq b (I64(Int64.zero))) then (I64(Int64.zero)) 
							else
								if (integerGt b (I64(Int64.zero))) then (I64(Int64.zero)) else (i64NextPowOfTwoNeg b nbIt)
|	(_,IMInfty) ->		 	IMInfty
|	(IMInfty,_) ->		 	IMInfty
|	(IPInfty,_)	-> 			b
|	(_,IPInfty)	-> 			a
|	(IBottom,_) ->			b
;;


let rec mpFloatNextPowOfTwoAux2 f n p = 
	if n<(-100) then 
		mpFloatPow (M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0)) n p GMP_RNDU
	else
		let w = mpFloatPow (M(FR.from_float_prec float32_mantissa GMP_RNDN 2.0)) n p GMP_RNDU
		in 	if (mpFloatGt w f) then
		  		mpFloatNextPowOfTwoAux2 f (n-1) p 
			else 
				mpFloatPow (M(FR.from_float_prec float32_mantissa GMP_RNDN 2.0)) (n+1) p GMP_RNDU ;;


let rec mpFloatNextPowOfTwoAux f n p = 
	if (n>1000) then MPInfty
	else
		let w = mpFloatPow (M(FR.from_float_prec float32_mantissa GMP_RNDN 2.0)) n p GMP_RNDU
		in 	if (mpFloatLt w f) then
		  		mpFloatNextPowOfTwoAux f (n+1) p 
			else w ;;
			

let mpFloatNextPowOfTwo f p nbIt = if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then MPInfty else match f with
	M(x) -> if ((FR.compare x (FR.from_float_prec float64_mantissa GMP_RNDN 1.0))>0) then
				mpFloatNextPowOfTwoAux f  (int_of_float ((log (getFloat f))/.(log 2.0))) p 
			else
				mpFloatNextPowOfTwoAux2 f 0 p
|   MPInfty -> MPInfty
|   MMInfty -> MMInfty
|   MBottom -> MBottom ;;
 

let mpFloatNextPowOfTwoNeg f p nbIt = if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then MMInfty else match f with
	M(x) ->	if ((FR.compare x (FR.from_float_prec float64_mantissa GMP_RNDN (-1.0)))<0) then
				let f' = mpFloatNeg f p GMP_RNDD in
				let w = mpFloatNextPowOfTwoAux f' (int_of_float ((log (getFloat f))/.(log 2.0))) p
				in mpFloatNeg w p GMP_RNDD
			else
				let f' = mpFloatNeg f p GMP_RNDD in
				let w = mpFloatNextPowOfTwoAux2 f' 0 p
				in mpFloatNeg w p GMP_RNDD
|   MPInfty -> MPInfty
|   MMInfty -> MMInfty
|   MBottom -> MBottom ;;


let mpFloatWidenRight a b p nbIt =  
	if (mpFloatEq a MBottom) then b 
	else
		if ((mpFloatLte a (M(FR.from_float_prec p GMP_RNDN 0.0)))&&
			(mpFloatEq  b (M(FR.from_float_prec p GMP_RNDN 0.0)))) 
		then 
			(M(FR.from_float_prec p GMP_RNDN 0.0)) 
		else  
			if (mpFloatLte b a) then a 
			else 
				if (mpFloatLt b (M(FR.from_float_prec p GMP_RNDN 0.0))) 											
				then 
					(M(FR.from_float_prec p GMP_RNDN 0.0)) 
			else (mpFloatNextPowOfTwo b p nbIt) ;;


let mpFloatWidenLeft a b p nbIt = 
  if (mpFloatEq a MBottom) then b 
  else
    if ((mpFloatGte a (M(FR.from_float_prec p GMP_RNDN 0.0)))&&
        (mpFloatEq  b (M(FR.from_float_prec p GMP_RNDN 0.0)))) 
    then 
      (M(FR.from_float_prec p GMP_RNDN 0.0)) 
    else
      if (mpFloatGte b a) then a 
	  else 
		if (mpFloatGt b (M(FR.from_float_prec p GMP_RNDN 0.0))) 
			then 
				(M(FR.from_float_prec p GMP_RNDN 0.0)) 
		else (mpFloatNextPowOfTwoNeg b p nbIt) ;;


let rec fixedNextPowOfTwoAux2 x n fp = 
	if n<(-126) then (xZero 1 fp)
	else
		let (w,_) = fixedPow (X(false,"0","1"^(String.make (fp-1) '0'))) n 1 fp 
		in 	if (fixedGt w x) then
				fixedNextPowOfTwoAux2 x (n+1) fp
			else
			 	fst (fixedPow (X(false,"0","1"^(String.make (fp-1) '0'))) (n-1) 1 fp) ;;


let rec fixedNextPowOfTwoAux x n fp = 
	if n>127 then XPInfty
	else
		let (w,ovf) = fixedPow (X(false,"10",(String.make fp '0'))) n (n+1) fp 
		in if ovf then XPInfty
			else
			 	if (fixedLt w x) then
		  			fixedNextPowOfTwoAux x (n+1) fp 
				else w ;;
			

let fixedNextPowOfTwo x fp nbIt = if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then XPInfty else match x with	
	X(_) -> 	if (fixedGt x (xOne 1 1)) then
					fixedNextPowOfTwoAux x (int_of_float ((log (getFloat (floatOfFixed x GMP_RNDU)))/.(log 2.0))) fp 
				else
					fixedNextPowOfTwoAux2 x 0 fp 
|   XPInfty -> XPInfty
|   XMInfty -> XMInfty
|   XBottom -> XBottom ;;


let fixedNextPowOfTwoNeg x fp nbIt = if ((nbIt-((!widenLevel)*(!unrollLevel)))>(!stopWiden)) then XMInfty else match x with
	X(_) -> if (fixedLt x (fixedNeg (xOne 1 1))) then
				let x' = fixedNeg x in
				let w = fixedNextPowOfTwoAux x' (int_of_float ((log (getFloat (floatOfFixed x GMP_RNDU)))/.(log 2.0))) fp
				in fixedNeg w
			else
				let x' = fixedNeg x in
				let w = fixedNextPowOfTwoAux2 x' 0 fp
				in fixedNeg w				 
|   XPInfty -> XPInfty
|   XMInfty -> XMInfty
|   XBottom -> XBottom ;;


let fixedWidenRight a b ip fp nbIt = 
	if (fixedEq a XBottom) then b 
	else
		if ((fixedLte a (xZero ip fp)) &&
			(fixedEq  b (xZero ip fp))) 
		then 
			(xZero ip fp) 
		else  
			if (fixedLte b a) then a 
			else 
				if (fixedLt b (xZero ip fp)) 											
				then 
					(xZero ip fp) 
			else fixedNextPowOfTwo b fp nbIt ;;


let fixedWidenLeft a b ip fp nbIt =
  if (fixedEq a XBottom) then b 
  else
    if ((fixedGte a (xZero ip fp)) &&
        (fixedEq  b (xZero ip fp))) 
    then 
      (xZero ip fp) 
    else
      if (fixedGte b a) then a 
	  else 
		if (fixedGt b (xZero ip fp)) 
			then 
				(xZero ip fp) 
		else fixedNextPowOfTwoNeg b fp nbIt ;;


let bWiden a b = match a with
  B(x) -> (match b with
             B(y) -> if (x=y) then a else BTop
           | BTop -> BTop
           | BBottom -> a
          )             
| BTop -> BTop
| BBottom  -> b ;;
  

(*****************************************************************************)
(*                  Widening on values and environments                      *)
(*****************************************************************************)


let iWiden a b nbIt = match (a,b) with
  ((v0Min,v0Max),(v1Min,v1Max)) -> ((integerWidenLeft v0Min v1Min nbIt,integerWidenRight v0Max v1Max nbIt)) ;;

 
let mpfiWiden a b p nbIt = match (a,b) with
  ((v0Min,v0Max),(v1Min,v1Max)) -> ((mpFloatWidenLeft v0Min v1Min p nbIt,mpFloatWidenRight v0Max v1Max p nbIt)) ;;


let fixedIntWiden a b ip fp nbIt = match (a,b) with
  ((v0Min,v0Max),(v1Min,v1Max)) -> ((fixedWidenLeft v0Min v1Min ip fp nbIt,fixedWidenRight v0Max v1Max ip fp nbIt)) ;;

let feWiden a b nbIt = match (a,b) with
	(FE(f1,e1,p1),FE(f2,e2,p2)) -> 	let p = max p1 p2
									in FE(mpfiWiden f1 f2 p nbIt,mpfiWiden e1 e2 !errPrec nbIt,p) ;;


let fiWiden a b nbIt = match (a,b) with
	(FX(f1,ip1,fp1,e1),FX(f2,ip2,fp2,e2)) -> 
		let ip = max ip1 ip2 in
		let fp = max ip2 fp2 in
			FX(fixedIntWiden f1 f2 ip fp nbIt,ip,fp,mpfiWiden e1 e2 !errPrec nbIt) ;;


 
let rec valWiden v1 v2 nbIt = match (v1,v2) with
	(IVal(x),IVal(y)) -> 				IVal(iWiden x y nbIt)
|	(FVal(x),FVal(y)) -> 				FVal(feWiden x y nbIt) 
|	(XVal(x),XVal(y)) ->			 	XVal(fiWiden x y nbIt)
| 	(BVal(x),BVal(y)) ->			 	BVal(bWiden x y)
| 	(IVal(_),FVal(_)) ->				valWiden (cast v1 (typeOfPrec (getFEPrec v2))) v2 nbIt
| 	(FVal(_),IVal(_)) -> 				valWiden v1 (cast v2 (typeOfPrec (getFEPrec v1))) nbIt
| 	(IVal(_),XVal(FX(_,ip,fp,_))) ->	valWiden (cast v1 (Type_Fixed(ip,fp))) v2 nbIt
| 	(XVal(FX(_,ip,fp,_)),IVal(_)) ->	valWiden v1 (cast v2 (Type_Fixed(ip,fp))) nbIt
|   (XVal(_),FVal(_)) -> 				raise (Error ("valWiden","??? Cannot corce fixed and float."))
|   (FVal(_),XVal(_)) -> 				raise (Error ("valWiden","??? Cannot corce fixed and float."))
|	(AVal(a1),AVal(a2)) -> 				let _ = checkArrayDims a1 a2  in
										let b = copyArray a1 in
										let _ = 
											for i=0 to ((Array.length b)-1) do
												for j=0 to ((Array.length b.(i))-1) do
													b.(i).(j) <- valWiden a1.(i).(j) a2.(i).(j) nbIt
												done
											done
										in AVal(b) 
|	(_,AVal(a)) ->						let b = copyArray a in 
										let _ = 
											for i=0 to ((Array.length b)-1) do
												for j=0 to ((Array.length b.(i))-1) do
													b.(i).(j) <- valWiden v1 a.(i).(j) nbIt
												done
											done
										in AVal(b) 							
|	(AVal(a),_) ->					 	let b = copyArray a in
										let _ = 
											for i=0 to ((Array.length b)-1) do
												for j=0 to ((Array.length b.(i))-1) do
													b.(i).(j) <- valWiden a.(i).(j) v2 nbIt
												done
											done
										in AVal(b) 							
|	(BVal(_),_) -> 						raise (Error ("valWiden","??? Type mismatch in widening (boolean not expected)."))
|	(_,BVal(_)) -> 						raise (Error ("valWiden","??? Type mismatch in widening (boolean not expected)."))
| 	(ValBottom,_) -> 					v2
| 	(_,ValBottom) -> 					v1
| 	(_,ValTop) ->						ValTop
|	(ValTop,_) ->						ValTop
|   (ArrayLines(_),_) ->			 	raise (Error ("valWiden","??? ArrayLines not expected in valWiden."))
|   (_,ArrayLines(_)) ->			 	raise (Error ("valWiden","??? ArrayLines not expected in valWiden."))
|   (ArrayColumns(_),_) -> 				raise (Error ("valWiden","??? ArrayColumns not expected in valWiden."))
|   (_,ArrayColumns(_)) -> 				raise (Error ("valWiden","??? ArrayColumns not expected in valWiden."))
;;


let rec envWiden env1 env2 nbIt = match env1 with
	[] -> 				env2
|	(id,v)::env1' -> let e = valWiden v (getEnvComplete id env2 ValBottom) nbIt
					 in envWiden env1' (setEnv env2 id e) nbIt ;;



